#pragma OPENCL EXTENSION cl_khr_byte_addressable_store : enable
#pragma OPENCL EXTENSION cl_khr_local_int32_base_atomics : enable
#pragma OPENCL EXTENSION cl_khr_global_int32_base_atomics : enable

// program constants
#define ALPHA (-1.586134342f)
#define BETA  (-0.05298011854f)
#define GAMMA (0.8829110762f)
#define DELTA (0.4435068522f)
#define ZETA1 (1.1496043988602418f)
#define ZETA2 (0.8697465511709978f)

// code selection enums
#define PROC_TYPE_BLAZ_NORMAL 0
#define PROC_TYPE_BLAZ_REGISTER 1
#define PROC_TYPE_LAAN 2


#ifndef ATOMIC_WARP
 #define ATOMIC_WARP 1
#endif

#if ATOMIC_WARP != 0
 #define WARP_SIZE 1
#endif

#ifndef WARP_SIZE
  #if DEVICE_TYPE == DEVICE_TYPE_GPU
    #define WARP_SIZE 32
  #else
    #define WAPR_SIZE 1
  #endif
#endif

#ifndef LOCAL_MEM_BANK_COUNT
  #define LOCAL_MEM_BANK_COUNT 32
#endif

#define DEVICE_TYPE_GPU 0
#define DEVICE_TYPE_CPU 1

#ifndef DEVICE_TYPE
#define DEVICE_TYPE DEVICE_TYPE_GPU
#endif


#ifndef VERT_BLOCK_SIZE
  #define VERT_BLOCK_SIZE 256
#endif

#ifndef VERT_PAIRS_PER_THREAD
  #define VERT_PAIRS_PER_THREAD 12
#endif

#ifndef VERT_BLOCK_SIZE_X
  #define VERT_BLOCK_SIZE_X 32
#endif

#define VERT_PIXELS_PER_THREAD (VERT_PAIRS_PER_THREAD * 2)

#define VERT_BLOCK_SIZE_Y (VERT_BLOCK_SIZE / VERT_BLOCK_SIZE_X)

#define VERT_PAIRS_PER_GROUP (VERT_BLOCK_SIZE_Y * VERT_PAIRS_PER_THREAD)

#define VERT_PIXELS_PER_GROUP (VERT_PAIRS_PER_GROUP * 2)

#define VERT_LOCAL_SIZE_Y (VERT_PIXELS_PER_GROUP + 7)

#define VERT_LOCAL_SIZE (VERT_BLOCK_SIZE_X * VERT_LOCAL_SIZE_Y)

#ifndef VERT_PROC_TYPE
  #define VERT_PROC_TYPE PROC_TYPE_BLAZ_NORMAL
#endif


#ifndef HOR_BLOCK_SIZE
  #define HOR_BLOCK_SIZE 256
#endif

#ifndef HOR_PAIRS_PER_THREAD
  #define HOR_PAIRS_PER_THREAD 1
#endif

#ifndef HOR_BLOCK_SIZE_Y
  #define HOR_BLOCK_SIZE_Y 1
#endif

#define HOR_PIXELS_PER_THREAD (HOR_PAIRS_PER_THREAD * 2)

#define HOR_BLOCK_SIZE_X (HOR_BLOCK_SIZE / HOR_BLOCK_SIZE_Y)

#define HOR_PAIRS_PER_GROUP (HOR_PAIRS_PER_THREAD * HOR_BLOCK_SIZE_X)

#define HOR_PIXELS_PER_GROUP (HOR_PAIRS_PER_GROUP * 2)

#define HOR_LOCAL_SIZE_X (HOR_PIXELS_PER_GROUP + 7)

#define HOR_LOCAL_SIZE (HOR_BLOCK_SIZE_Y * HOR_LOCAL_SIZE_X)

#ifndef HOR_PROC_TYPE
  #define HOR_PROC_TYPE PROC_TYPE_BLAZ_NORMAL
#endif

#if DEVICE_TYPE == DEVICE_TYPE_GPU
  #define HOR_LOCAL_SIZE_ALIGN (HOR_LOCAL_SIZE + (HOR_LOCAL_SIZE / LOCAL_MEM_BANK_COUNT))
  size_t change_offset(size_t offset)
    {
      return offset + (offset >> 5);
    }
#else
  #define HOR_LOCAL_SIZE_ALIGN HOR_LOCAL_SIZE
  inline size_t change_offset(size_t offset)
    {
      return offset; 
    }
#endif

#if (WARP_SIZE >= HOR_BLOCK_SIZE_X) && ((WARP_SIZE % HOR_BLOCK_SIZE_X) == 0)
#define HOR_ATOM 1
#else
#define HOR_ATOM 0
#endif

void load_vert_t(__global float *in_image, __local float *temp_image, size_t ly, uint stride)
{
  float act_val = in_image[0];
  temp_image[4 * VERT_BLOCK_SIZE_X] = act_val;
  if((ly < 5) && (ly != 0))
    {
      temp_image[(4 - 2 * ly) * VERT_BLOCK_SIZE_X] = act_val;
    }
  for(int i = 1; i < VERT_PIXELS_PER_THREAD; i++)
    {
      temp_image[(4 + i * VERT_BLOCK_SIZE_Y) * VERT_BLOCK_SIZE_X] = in_image[(i * VERT_BLOCK_SIZE_Y) * stride];
    }
  if(ly < 3)
    {
      temp_image[(4 + VERT_PIXELS_PER_GROUP) * VERT_BLOCK_SIZE_X] = in_image[VERT_PIXELS_PER_GROUP * stride];
    }
}

void load_vert_m(__global float *in_image, __local float *temp_image, size_t ly, uint stride)
{
  // copy previous data
  if(ly < 7)
    {
      temp_image[0] = temp_image[(VERT_LOCAL_SIZE_Y - 7) * VERT_BLOCK_SIZE_X];
    }
  // get next data
  for(int i = 0; i < VERT_PIXELS_PER_THREAD; i++)
    {
      temp_image[(7 + i * VERT_BLOCK_SIZE_Y) * VERT_BLOCK_SIZE_X] = in_image[(i * VERT_BLOCK_SIZE_Y) * stride];
    }
}

void load_vert_b(__global float *in_image, __local float *temp_image, size_t ly, uint stride)
{
  float act_val;
  // copy previous data
  if(ly < 7)
    {
      temp_image[0] = temp_image[(VERT_LOCAL_SIZE_Y - 7) * VERT_BLOCK_SIZE_X];
    }
  // get next data
  for(int i = 0; i < VERT_PIXELS_PER_THREAD - 1; i++)
    {
      temp_image[(7 + i * VERT_BLOCK_SIZE_Y) * VERT_BLOCK_SIZE_X] = in_image[(i * VERT_BLOCK_SIZE_Y) * stride];
    }
  if(ly < VERT_BLOCK_SIZE_Y - 3)
    {
      act_val = in_image[(VERT_PIXELS_PER_GROUP - VERT_BLOCK_SIZE_Y) * stride];
      temp_image[(7 + VERT_PIXELS_PER_GROUP - VERT_BLOCK_SIZE_Y) * VERT_BLOCK_SIZE_X] = act_val;
    }

  if((ly < VERT_BLOCK_SIZE_Y - 4) && (ly >= VERT_BLOCK_SIZE_Y - 7))
    {
      temp_image[(4 + VERT_PIXELS_PER_GROUP + VERT_BLOCK_SIZE_Y - 5 - 2 * ly) * VERT_BLOCK_SIZE_X] = act_val;
    }
}

void load_vert_m_split(__global float *in_image, __local float *temp_image, size_t ly, uint stride)
{
  // get next data
  for(int i = 0; i < VERT_PIXELS_PER_THREAD; i++)
    {
      temp_image[(i * VERT_BLOCK_SIZE_Y) * VERT_BLOCK_SIZE_X] = in_image[(i * VERT_BLOCK_SIZE_Y) * stride];
    }
  // copy previous data
  if(ly < 7)
    {
      temp_image[VERT_PIXELS_PER_GROUP * VERT_BLOCK_SIZE_X] = in_image[VERT_PIXELS_PER_GROUP * stride];
    }
}

void load_vert_b_split(__global float *in_image, __local float *temp_image, size_t ly, uint stride)
{
  float act_val;
  for(int i = 0; i < VERT_PIXELS_PER_THREAD; i++)
    {
	  temp_image[(i * VERT_BLOCK_SIZE_Y) * VERT_BLOCK_SIZE_X] = in_image[(i * VERT_BLOCK_SIZE_Y) * stride];
	}
  if(ly < 4)
	{
	  act_val = in_image[VERT_PIXELS_PER_GROUP * stride];
	  temp_image[VERT_PIXELS_PER_GROUP * VERT_BLOCK_SIZE_X] = act_val;
	}
  if(ly < 3)
	{
	  temp_image[(6 + VERT_PIXELS_PER_GROUP - ly * 2) * VERT_BLOCK_SIZE_X] = act_val;
	}
}

void load_vert_b_doublebuffer(__global float *in_image, __local float *temp_image, __local float *temp_image_old, size_t ly, uint stride)
{
  float act_val;
  // copy previous data
  if(ly < 7)
    {
      temp_image[0] = temp_image_old[(VERT_LOCAL_SIZE_Y - 7) * VERT_BLOCK_SIZE_X];
    }
  // get next data
  for(int i = 0; i < VERT_PIXELS_PER_THREAD - 1; i++)
    {
      temp_image[(7 + i * VERT_BLOCK_SIZE_Y) * VERT_BLOCK_SIZE_X] = in_image[(i * VERT_BLOCK_SIZE_Y) * stride];
    }
  if(ly < VERT_BLOCK_SIZE_Y - 3)
    {
      act_val = in_image[(VERT_PIXELS_PER_GROUP - VERT_BLOCK_SIZE_Y) * stride];
      temp_image[(7 + VERT_PIXELS_PER_GROUP - VERT_BLOCK_SIZE_Y) * VERT_BLOCK_SIZE_X] = act_val;
    }

  if((ly < VERT_BLOCK_SIZE_Y - 4) && (ly >= VERT_BLOCK_SIZE_Y - 7))
    {
      temp_image[(4 + VERT_PIXELS_PER_GROUP + VERT_BLOCK_SIZE_Y - 5 - 2 * ly) * VERT_BLOCK_SIZE_X] = act_val;
    }
}

void load_vert_m_doublebuffer(__global float *in_image, __local float *temp_image, __local float *temp_image_old, size_t ly, uint stride)
{
  // copy previous data
  if(ly < 7)
    {
      temp_image[0] = temp_image_old[(VERT_LOCAL_SIZE_Y - 7) * VERT_BLOCK_SIZE_X];
    }
  // get next data
  for(int i = 0; i < VERT_PIXELS_PER_THREAD; i++)
    {
      temp_image[(7 + i * VERT_BLOCK_SIZE_Y) * VERT_BLOCK_SIZE_X] = in_image[(i * VERT_BLOCK_SIZE_Y) * stride];
    }
}


#if VERT_PROC_TYPE == PROC_TYPE_BLAZ_NORMAL
void proc_vert_blazewicz(__local float *local_image, __global float *out_image_l, __global float *out_image_h, uint stride)
{
  // create private memory
  __private float private_image[VERT_PIXELS_PER_THREAD + 5];
  for (int i = 0; i < VERT_PIXELS_PER_THREAD + 5; i++)
    {
      private_image[i] = local_image[(i + 1) * VERT_BLOCK_SIZE_X];
    }
  // compute wavelet transform
  private_image[0] += ALPHA * (private_image[1] + local_image[0]);
  private_image[VERT_PIXELS_PER_THREAD + 4] += ALPHA * (private_image[VERT_PIXELS_PER_THREAD + 3] + local_image[(VERT_PIXELS_PER_THREAD + 6) * VERT_BLOCK_SIZE_X]);
  for(int i = 1; i < VERT_PAIRS_PER_THREAD + 2; i++)
    {
      private_image[i * 2] += ALPHA * (private_image[i * 2 - 1] + private_image[i * 2 + 1]);
    }
  for(int i = 0; i < VERT_PAIRS_PER_THREAD + 2; i++)
    {
      private_image[i * 2 + 1] += BETA * (private_image[i * 2] + private_image[i * 2 + 2]);
    }
  for(int i = 1; i < VERT_PAIRS_PER_THREAD + 2; i++)
    {
      private_image[i * 2] += GAMMA * (private_image[i * 2 - 1] + private_image[i * 2 + 1]);
    }
  for(int i = 1; i < VERT_PAIRS_PER_THREAD + 1; i++)
    {
      private_image[i * 2 + 1] += DELTA * (private_image[i * 2] + private_image[i * 2 + 2]);
    }
  // save output coefficients
  for(int i = 0; i < VERT_PAIRS_PER_THREAD; i++)
    {
      out_image_l[i * stride] = ZETA1 * private_image[i * 2 + 3];
      out_image_h[i * stride] = ZETA2 * private_image[i * 2 + 4];
    }
}
#else
void proc_vert_blazewicz(__local float *local_image, __global float *out_image_l, __global float *out_image_h, uint stride)
{
  // create private memory
  __private float private_image[5];
  float tmp_image_0, tmp_image_1, tmp_image_2;
  
  private_image[0] = local_image[0];

  tmp_image_0 = private_image[0];
  private_image[0] = local_image[2 * VERT_BLOCK_SIZE_X];
  private_image[1] = local_image[1 * VERT_BLOCK_SIZE_X] + ALPHA * (tmp_image_0 + private_image[0]);

  tmp_image_0 = private_image[0];
  tmp_image_1 = private_image[1];
  private_image[0] = local_image[4 * VERT_BLOCK_SIZE_X];
  private_image[1] = local_image[3 * VERT_BLOCK_SIZE_X] + ALPHA * (tmp_image_0 + private_image[0]);
  private_image[2] = tmp_image_0 + BETA * (tmp_image_1 + private_image[1]);
  
  tmp_image_0 = private_image[0];
  tmp_image_1 = private_image[1];
  tmp_image_2 = private_image[2];
  private_image[0] = local_image[6 * VERT_BLOCK_SIZE_X];
  private_image[1] = local_image[5 * VERT_BLOCK_SIZE_X] + ALPHA * (tmp_image_0 + private_image[0]);
  private_image[2] = tmp_image_0 + BETA * (tmp_image_1 + private_image[1]);
  private_image[3] = tmp_image_1 + GAMMA * (tmp_image_2 + private_image[2]);

  for(int i = 0; i < VERT_PAIRS_PER_THREAD; i++)
    {
      tmp_image_0 = private_image[0];
      tmp_image_1 = private_image[1];
      tmp_image_2 = private_image[2];
      private_image[0] = local_image[(8 + i * 2) * VERT_BLOCK_SIZE_X];
      private_image[1] = local_image[(7 + i * 2) * VERT_BLOCK_SIZE_X] + ALPHA * (tmp_image_0 + private_image[0]);
      private_image[2] = tmp_image_0 + BETA * (tmp_image_1 + private_image[1]);
      tmp_image_0 = private_image[3];
      private_image[3] = tmp_image_1 + GAMMA * (tmp_image_2 + private_image[2]);
      private_image[4] = tmp_image_2 + DELTA * (tmp_image_0 + private_image[3]);
      out_image_l[i * stride] = ZETA1 * private_image[4];
      out_image_h[i * stride] = ZETA2 * private_image[3];
    }
}
#endif
// requirements:
//  width align to VERT_BLOCK_SIZE_X
//  height align to VERT_BLOCK_SIZE_Y * PAIR_PER_THREAD * 2
//  minimum height 2 * VERT_BLOCK_SIZE_Y * PAIR_PER_THREAD * 2
#ifdef wavelet_vert_basic
__kernel void wavelet_vert(__global float *out_image, __global float *in_image, uint width, uint height, uint depth)
{
  size_t lx = get_local_id(0);
  size_t ly = get_local_id(1);

  #ifndef WAVELET_OUTPUT_INTERLACED
	  size_t gx = get_global_id(0);
	  size_t gy = get_global_id(1);
    int gy_l = gy * VERT_PAIRS_PER_THREAD;
    int gy_l_to_h_stride = height>>(1+depth);
    int gy_out_stride = 1;
    int gy_block_stride = VERT_PAIRS_PER_GROUP;
	  int gy_in_stride = 1;
  #else
	  size_t gx = get_global_id(0) << depth;
	  size_t gy = get_global_id(1) << depth;
    int gy_l = gy * VERT_PIXELS_PER_THREAD;
    int gy_l_to_h_stride = 1 << depth;
    int gy_out_stride = 1 << (1 + depth);
    int gy_block_stride = VERT_PIXELS_PER_GROUP << depth;
	  int gy_in_stride = 1 << depth;
  #endif

  int out_stride = gy_out_stride * width;
  int l_to_h_stride = gy_l_to_h_stride * width;
  int block_stride = gy_block_stride * width;
  int in_stride = gy_in_stride * width;

  __local float temp_image[VERT_LOCAL_SIZE];

  __global float *g_in_image = in_image + gx + gy * width;
  __global float *g_out_image_l = out_image + gx + gy_l * width;
  __global float *g_out_image_h = g_out_image_l + l_to_h_stride;
  __local float *l_load_image_pos = temp_image + lx + ly * VERT_BLOCK_SIZE_X;
  __local float *l_proc_image_pos = temp_image + lx + ly * VERT_PIXELS_PER_THREAD * VERT_BLOCK_SIZE_X;
  
  load_vert_t(g_in_image, l_load_image_pos, ly, in_stride);
  barrier(CLK_LOCAL_MEM_FENCE);
  proc_vert_blazewicz(l_proc_image_pos, g_out_image_l, g_out_image_h, out_stride);
  g_in_image += (VERT_PIXELS_PER_GROUP + 3) * in_stride;
  g_out_image_l += block_stride;
  g_out_image_h += block_stride;
  for(int i = 1; i < ((height>>depth) / VERT_PIXELS_PER_GROUP) - 1; i++)
    {
      barrier(CLK_LOCAL_MEM_FENCE);
      load_vert_m(g_in_image, l_load_image_pos, ly, in_stride);
      barrier(CLK_LOCAL_MEM_FENCE);
      proc_vert_blazewicz(l_proc_image_pos, g_out_image_l, g_out_image_h, out_stride);
      g_in_image += VERT_PIXELS_PER_GROUP * in_stride;
      g_out_image_l += block_stride;
	  g_out_image_h += block_stride;
    }
  barrier(CLK_LOCAL_MEM_FENCE);
  load_vert_b(g_in_image, l_load_image_pos, ly, in_stride);
  barrier(CLK_LOCAL_MEM_FENCE);
  proc_vert_blazewicz(l_proc_image_pos, g_out_image_l, g_out_image_h, out_stride);
}
#endif


#ifdef wavelet_vert_split
__kernel void wavelet_vert(__global float *out_image, __global float *in_image, uint width, uint height, uint depth)
{
  size_t lx = get_local_id(0);
  size_t ly = get_local_id(1);
  size_t gry = get_group_id(1);
	
  #ifndef WAVELET_OUTPUT_INTERLACED
	  size_t gx = get_global_id(0);
	  size_t gy = get_global_id(1);
    int gy_l = gy * VERT_PAIRS_PER_THREAD;
    int gy_l_to_h_stride = height>>(1+depth);
    int gy_out_stride = 1;
	  int gy_in_stride = 1;
  #else
	  size_t gx = get_global_id(0) << depth;
	  size_t gy = get_global_id(1) << depth;
    int gy_l = gy * VERT_PIXELS_PER_THREAD;
    int gy_l_to_h_stride = 1 << depth;
    int gy_out_stride = 1 << (1 + depth);
	int gy_in_stride = 1 << depth;
  #endif

  int out_stride = gy_out_stride * width;
  int l_to_h_stride = gy_l_to_h_stride * width;
  int in_stride = gy_in_stride * width;

  __local float temp_image[VERT_LOCAL_SIZE];

  __global float *g_in_image = in_image + gx + ly * in_stride;
  __global float *g_out_image_l = out_image + gx + gy_l * width;
  __global float *g_out_image_h = g_out_image_l + l_to_h_stride;
  __local float *l_load_image_pos = temp_image + lx + ly * VERT_BLOCK_SIZE_X;
  __local float *l_proc_image_pos = temp_image + lx + ly * VERT_PIXELS_PER_THREAD * VERT_BLOCK_SIZE_X;
  if(gry == 0)
    {
      load_vert_t(g_in_image, l_load_image_pos, ly, in_stride);
	    barrier(CLK_LOCAL_MEM_FENCE);
	    proc_vert_blazewicz(l_proc_image_pos, g_out_image_l, g_out_image_h, out_stride);
    }
  else if(gry != get_num_groups(1) - 1)
    {
      g_in_image += (VERT_PIXELS_PER_GROUP * gry - 4) * in_stride;
      load_vert_m_split(g_in_image, l_load_image_pos, ly, in_stride);
      barrier(CLK_LOCAL_MEM_FENCE);
      proc_vert_blazewicz(l_proc_image_pos, g_out_image_l, g_out_image_h, out_stride);
    }
  else
    {
      g_in_image += (VERT_PIXELS_PER_GROUP * gry - 4) * in_stride;
      load_vert_b_split(g_in_image, l_load_image_pos, ly, in_stride);
      barrier(CLK_LOCAL_MEM_FENCE);
      proc_vert_blazewicz(l_proc_image_pos, g_out_image_l, g_out_image_h, out_stride);
    }
}
#endif


#ifdef wavelet_vert_line
__kernel void wavelet_vert(__global float *out_image, __global float *in_image, uint width, uint height, uint depth)
{ 
  #ifndef WAVELET_OUTPUT_INTERLACED
	  size_t gx = get_global_id(0);
	  size_t gy = get_global_id(1);
    size_t gy_l = gy * VERT_PAIRS_PER_THREAD;
    int gy_l_to_h = height>>(1+depth);
    int gy_out_stride = 1;
	  int gy_in_stride = 1;
  #else
	  size_t gx = get_global_id(0) << depth;
	  size_t gy = get_global_id(1) << depth;
    size_t gy_l = gy * VERT_PIXELS_PER_THREAD;
    int gy_l_to_h = 1 << depth;
    int gy_out_stride = 1 << (1 + depth);
	  int gy_in_stride = 1 << depth;
  #endif

  size_t gy_in = gy * VERT_PIXELS_PER_THREAD;

  int out_stride = gy_out_stride * width;
  int in_stride = gy_in_stride * width;

  __global float *g_in_image = in_image + gx + gy_in * width;
  __global float *g_out_image_l = out_image + gx + gy_l * width;
  __global float *g_out_image_h = g_out_image_l + gy_l_to_h * width;

  float a,b,c,d,e,tmp1;
  if(gy == 0)
    {
      e = *g_in_image;
      g_in_image += in_stride;
      d = *g_in_image;
      g_in_image += in_stride;
      c = *g_in_image;
      g_in_image += in_stride;

      d += ALPHA * (e + c);
      e += BETA * 2 * d;

      b = *g_in_image;
      g_in_image += in_stride;
      a = *g_in_image;
      g_in_image += in_stride;

      
      b += ALPHA * (c + a);
      c += BETA * (d + b);
      d += GAMMA * (e + c);
      e += DELTA * 2 * d;
    }
  else
    {
      g_in_image -= in_stride * 4;
      
      e = *g_in_image;
      g_in_image += in_stride;
      tmp1 = *g_in_image;
      g_in_image += in_stride;
      d = *g_in_image;
      g_in_image += in_stride;

      e = tmp1 + ALPHA * (e + d);
      
      tmp1 = *g_in_image;
      g_in_image += in_stride;
      c = *g_in_image;
      g_in_image += in_stride;
      
      e = d + BETA * e;
      d = tmp1 + ALPHA * (d + c);
      e += BETA * d;

      tmp1 = *g_in_image;
      g_in_image += in_stride;
      b = *g_in_image;
      g_in_image += in_stride;

      e = d + GAMMA * e;
      d = c + BETA * d;
      c = tmp1 + ALPHA * (c + b);
      d += BETA * c;
      e += GAMMA * d;

      tmp1 = *g_in_image;
      g_in_image += in_stride;
      a = *g_in_image;
      g_in_image += in_stride;

      e = d + DELTA * e;
      d = c + GAMMA * d;
      c = b + BETA * c;
      b = tmp1 + ALPHA * (b + a);
      c += BETA * b;
      d += GAMMA * c;
      e += DELTA * d;
    }
  *g_out_image_l = ZETA1 * e;
  *g_out_image_h = ZETA2 * d;
  g_out_image_l += out_stride;
  g_out_image_h += out_stride;

  for(int i = 1; i < VERT_PAIRS_PER_THREAD - 2; i++)
    {
      tmp1 = *g_in_image;
      g_in_image += in_stride;

      e = DELTA * d + c;
      d = GAMMA * c + b;
      c = BETA * b + a;
      b = ALPHA * a + tmp1;

      a = *g_in_image;
      g_in_image += in_stride;

      b += ALPHA * a;
      c += BETA * b;
      d += GAMMA * c;
      e += DELTA * d;

      *g_out_image_l = ZETA1 * e;
      *g_out_image_h = ZETA2 * d;
      g_out_image_l += out_stride;
      g_out_image_h += out_stride;
    }
  
  if(gy == get_global_size(1) - 1)
    {
      tmp1 = *g_in_image;
      g_in_image += in_stride;

      e = DELTA * d + c;
      d = GAMMA * c + b;
      c = BETA * b + a;
      b = ALPHA * 2 * a + tmp1;
      c += BETA * b;
      d += GAMMA * c;
      e += DELTA * d;

      *g_out_image_l = ZETA1 * e;
      *g_out_image_h = ZETA2 * d;
      g_out_image_l += out_stride;
      g_out_image_h += out_stride;

      e = DELTA * d + c;
      d = GAMMA * 2 * c + b;
      e += DELTA * d;

      *g_out_image_l = ZETA1 * e;
      *g_out_image_h = ZETA2 * d;
    }
  else
    {
      tmp1 = *g_in_image;
      g_in_image += in_stride;

      e = DELTA * d + c;
      d = GAMMA * c + b;
      c = BETA * b + a;
      b = ALPHA * a + tmp1;

      a = *g_in_image;
      g_in_image += in_stride;

      b += ALPHA * a;
      c += BETA * b;
      d += GAMMA * c;
      e += DELTA * d;

      *g_out_image_l = ZETA1 * e;
      *g_out_image_h = ZETA2 * d;
      g_out_image_l += out_stride;
      g_out_image_h += out_stride;

      tmp1 = *g_in_image;
      g_in_image += in_stride;

      e = DELTA * d + c;
      d = GAMMA * c + b;
      c = BETA * b + a;
      b = ALPHA * a + tmp1;

      a = *g_in_image;
      g_in_image += in_stride;

      b += ALPHA * a;
      c += BETA * b;
      d += GAMMA * c;
      e += DELTA * d;

      *g_out_image_l = ZETA1 * e;
      *g_out_image_h = ZETA2 * d;
    }
}
#endif

void load_hor_l(__global float *in_image, __local float *temp_image, size_t lx, size_t l_in_image_pos, int in_stride)
{
  float act_val = in_image[0];
  if((lx < 5) && (lx != 0))
    {
      temp_image[change_offset(l_in_image_pos + 4 - 2 * lx)] = act_val;
    }
  temp_image[change_offset(l_in_image_pos + 4)] = act_val;
  for(int i = 1; i < HOR_PIXELS_PER_THREAD; i++)
    {
      temp_image[change_offset(l_in_image_pos + 4 + i * HOR_BLOCK_SIZE_X)] = in_image[i * HOR_BLOCK_SIZE_X * in_stride];
    }
  if(lx < 3)
    {
      temp_image[change_offset(l_in_image_pos + 4 + HOR_PIXELS_PER_GROUP)] = in_image[HOR_PIXELS_PER_GROUP * in_stride];
    }
}

void load_hor_lr(__global float *in_image, __local float *temp_image, size_t lx, size_t l_in_image_pos, int in_stride)
{
  float act_val = in_image[0];
  if((lx < 5) && (lx != 0))
    {
      temp_image[change_offset(l_in_image_pos + 4 - 2 * lx)] = act_val;
    }
  temp_image[change_offset(l_in_image_pos + 4)] = act_val;
  for(int i = 1; i < HOR_PIXELS_PER_THREAD - 1; i++)
    {
      temp_image[change_offset(l_in_image_pos + 4 + i * HOR_BLOCK_SIZE_X)] = in_image[i * HOR_BLOCK_SIZE_X * in_stride];
    }
  act_val = in_image[(HOR_PIXELS_PER_GROUP - HOR_BLOCK_SIZE_X) * in_stride];
  temp_image[change_offset(l_in_image_pos + 4 + HOR_PIXELS_PER_GROUP - HOR_BLOCK_SIZE_X)] = act_val;
  if((lx >= HOR_BLOCK_SIZE_X - 4) && (lx != HOR_BLOCK_SIZE_X - 1))
    {
      temp_image[change_offset(l_in_image_pos + 2 + HOR_PIXELS_PER_GROUP + HOR_BLOCK_SIZE_X - 2 * lx)] = act_val;
    }
}

void load_hor_m(__global float *in_image, __local float *temp_image, size_t lx, size_t l_in_image_pos, int in_stride)
{
  // copy previous data
  if(lx < 7)
    {
      temp_image[change_offset(l_in_image_pos)] = temp_image[change_offset(l_in_image_pos + HOR_LOCAL_SIZE_X - 7)];
    }
  // get next data
  for(int i = 0; i < HOR_PIXELS_PER_THREAD; i++)
    {
      temp_image[change_offset(l_in_image_pos + 7 + i * HOR_BLOCK_SIZE_X)] = in_image[i * HOR_BLOCK_SIZE_X * in_stride];
    }
}

void load_hor_r(__global float *in_image, __local float *temp_image, size_t lx, size_t l_in_image_pos, int in_stride)
{
  float act_val;
  if(lx < 7)
    {
      temp_image[change_offset(l_in_image_pos)] = temp_image[change_offset(l_in_image_pos + HOR_LOCAL_SIZE_X - 7)];
    }
  for(int i = 0; i < HOR_PIXELS_PER_THREAD - 1; i++)
    {
      temp_image[change_offset(l_in_image_pos + 7 + i * HOR_BLOCK_SIZE_X)] = in_image[i * HOR_BLOCK_SIZE_X * in_stride];
    }
  if(lx < HOR_BLOCK_SIZE_X - 3)
    {
      act_val = in_image[(HOR_PIXELS_PER_THREAD - 1) * HOR_BLOCK_SIZE_X * in_stride];
      temp_image[change_offset(l_in_image_pos + 7 + (HOR_PIXELS_PER_THREAD - 1) * HOR_BLOCK_SIZE_X)] = act_val;
    }
  if((lx < HOR_BLOCK_SIZE_X - 4) && (lx >= HOR_BLOCK_SIZE_X - 7))
    {
      temp_image[change_offset(l_in_image_pos + 4 + HOR_PIXELS_PER_GROUP + HOR_BLOCK_SIZE_X - 5 - 2 * lx)] = act_val;
    }
}

void load_hor_m_split(__global float *in_image, __local float *temp_image, size_t lx, size_t l_in_image_pos, int in_stride)
{
  for(int i = 0; i < HOR_PIXELS_PER_THREAD; i++)
    {
      temp_image[change_offset(l_in_image_pos + i * HOR_BLOCK_SIZE_X)] = in_image[i * HOR_BLOCK_SIZE_X * in_stride];
    }
  // copy previous data
  if(lx < 7)
    {
      temp_image[change_offset(l_in_image_pos + HOR_PIXELS_PER_GROUP)] = in_image[HOR_PIXELS_PER_GROUP * in_stride];
    }
}

void load_hor_r_split(__global float *in_image, __local float *temp_image, size_t lx, size_t l_in_image_pos, int in_stride)
{
  float act_val;
  for(int i = 0; i < HOR_PIXELS_PER_THREAD; i++)
    {
      temp_image[change_offset(l_in_image_pos + i * HOR_BLOCK_SIZE_X)] = in_image[i * HOR_BLOCK_SIZE_X * in_stride];
    }
  if(lx < 4)
    {
      act_val = in_image[HOR_PIXELS_PER_GROUP * in_stride];
      temp_image[change_offset(l_in_image_pos + HOR_PIXELS_PER_GROUP)] = act_val;
    }
  if(lx < 3)
	{
      temp_image[change_offset(l_in_image_pos + 6 + HOR_PIXELS_PER_GROUP - 2 * lx)] = act_val;
	}
}

#if HOR_PROC_TYPE == PROC_TYPE_BLAZ_NORMAL
void proc_hor(__local float *l_in_image, __global float *out_image_l, __global float *out_image_h, size_t l_in_image_pos, size_t out_stride)
{
  // create private memory
  __private float private_image[HOR_PIXELS_PER_THREAD + 5];
  for(int i = 0; i < HOR_PIXELS_PER_THREAD + 5; i++)
    {
    private_image[i] = l_in_image[change_offset(l_in_image_pos + i + 1)];
    }
  // compute wavelet transform
  private_image[0] += ALPHA * (private_image[1] + l_in_image[change_offset(l_in_image_pos)]);
  private_image[HOR_PIXELS_PER_THREAD + 4] += ALPHA * (private_image[HOR_PIXELS_PER_THREAD + 3] + l_in_image[change_offset(l_in_image_pos + HOR_PIXELS_PER_THREAD + 6)]);
  for(int i = 1; i < HOR_PAIRS_PER_THREAD + 2; i++)
    {
      private_image[i * 2] += ALPHA * (private_image[i * 2 - 1] + private_image[i * 2 + 1]);
    }
  for(int i = 0; i < HOR_PAIRS_PER_THREAD + 2; i++)
    {
      private_image[i * 2 + 1] += BETA * (private_image[i * 2] + private_image[i * 2 + 2]);
    }
  for(int i = 1; i < HOR_PAIRS_PER_THREAD + 2; i++)
    {
      private_image[i * 2] += GAMMA * (private_image[i * 2 - 1] + private_image[i * 2 + 1]);
    }
  for(int i = 1; i < HOR_PAIRS_PER_THREAD + 1; i++)
    {
      private_image[i * 2 + 1] += DELTA * (private_image[i * 2] + private_image[i * 2 + 2]);
    }
  // save output coefficients
  for(int i = 0; i < HOR_PAIRS_PER_THREAD; i++)
    {
      out_image_l[i * out_stride] = ZETA1 * private_image[i * 2 + 3];
      out_image_h[i * out_stride] = ZETA2 * private_image[i * 2 + 4];
    }
}
#elif HOR_PROC_TYPE == PROC_TYPE_BLAZ_REGISTER
void proc_hor(__local float *l_in_image, __global float *out_image_l, __global float *out_image_h, size_t l_in_image_pos, size_t out_stride)
{
  // create private memory
  __private float private_image[5];
  float tmp_image_0, tmp_image_1, tmp_image_2;

  private_image[0] = l_in_image[change_offset(l_in_image_pos)];

  tmp_image_0 = private_image[0];
  private_image[0] = l_in_image[change_offset(l_in_image_pos + 2)];
  private_image[1] = l_in_image[change_offset(l_in_image_pos + 1)] + ALPHA * (tmp_image_0 + private_image[0]);

  tmp_image_0 = private_image[0];
  tmp_image_1 = private_image[1];
  private_image[0] = l_in_image[change_offset(l_in_image_pos + 4)];
  private_image[1] = l_in_image[change_offset(l_in_image_pos + 3)] + ALPHA * (tmp_image_0 + private_image[0]);
  private_image[2] = tmp_image_0 + BETA * (tmp_image_1 + private_image[1]);

  tmp_image_0 = private_image[0];
  tmp_image_1 = private_image[1];
  tmp_image_2 = private_image[2];
  private_image[0] = l_in_image[change_offset(l_in_image_pos + 6)];
  private_image[1] = l_in_image[change_offset(l_in_image_pos + 5)] + ALPHA * (tmp_image_0 + private_image[0]);
  private_image[2] = tmp_image_0 + BETA * (tmp_image_1 + private_image[1]);
  private_image[3] = tmp_image_1 + GAMMA * (tmp_image_2 + private_image[2]);

  for (int i = 0; i < HOR_PAIRS_PER_THREAD; i++)
  {
    tmp_image_0 = private_image[0];
    tmp_image_1 = private_image[1];
    tmp_image_2 = private_image[2];
    private_image[0] = l_in_image[change_offset(l_in_image_pos + 7 + i * 2)];
    private_image[1] = l_in_image[change_offset(l_in_image_pos + 8 + i * 2)] + ALPHA * (tmp_image_0 + private_image[0]);
    private_image[2] = tmp_image_0 + BETA * (tmp_image_1 + private_image[1]);
    tmp_image_0 = private_image[3];
    private_image[3] = tmp_image_1 + GAMMA * (tmp_image_2 + private_image[2]);
    private_image[4] = tmp_image_2 + DELTA * (tmp_image_0 + private_image[3]);
    out_image_l[i * out_stride] = ZETA1 * private_image[4];
    out_image_h[i * out_stride] = ZETA2 * private_image[3];
  }
}
#else
void proc_hor_start(__local float *l_in_image, size_t l_in_image_pos, size_t lx)
{
  if(lx < 3) l_in_image[change_offset(l_in_image_pos + 1)] += ALPHA * (l_in_image[change_offset(l_in_image_pos + 2)] + l_in_image[change_offset(l_in_image_pos)]);
  #if WARP_SIZE < 3
  barrier(CLK_LOCAL_MEM_FENCE);
  #endif
  
  if(lx < 2) l_in_image[change_offset(l_in_image_pos + 2)] += BETA * (l_in_image[change_offset(l_in_image_pos + 3)] + l_in_image[change_offset(l_in_image_pos + 1)]);
  #if WARP_SIZE < 2
  barrier(CLK_LOCAL_MEM_FENCE);
  #endif
  
  if(lx < 1) l_in_image[change_offset(l_in_image_pos + 3)] += GAMMA * (l_in_image[change_offset(l_in_image_pos + 4)] + l_in_image[change_offset(l_in_image_pos + 2)]);
}

void proc_hor(__local float *l_in_image, __global float *out_image_l, __global float *out_image_h, size_t l_in_image_pos, size_t out_stride)
{
  for(int i = 0; i < HOR_PAIRS_PER_THREAD; i++)
    {
      l_in_image[change_offset(l_in_image_pos + i * 2 + 7)] += ALPHA * (l_in_image[change_offset(l_in_image_pos + i * 2 + 8)] + l_in_image[change_offset(l_in_image_pos + i * 2 + 6)]);
    }
  #if HOR_ATOM == 0
  barrier(CLK_LOCAL_MEM_FENCE);
  #endif
  for(int i = 0; i < HOR_PAIRS_PER_THREAD; i++)
    {
      l_in_image[change_offset(l_in_image_pos + i * 2 + 6)] += BETA * (l_in_image[change_offset(l_in_image_pos + i * 2 + 7)] + l_in_image[change_offset(l_in_image_pos + i * 2 + 5)]);
    }
  #if HOR_ATOM == 0
  barrier(CLK_LOCAL_MEM_FENCE);
  #endif
  for(int i = 0; i < HOR_PAIRS_PER_THREAD; i++)
    {
      l_in_image[change_offset(l_in_image_pos + i * 2 + 5)] += GAMMA * (l_in_image[change_offset(l_in_image_pos + i * 2 + 6)] + l_in_image[change_offset(l_in_image_pos + i * 2 + 4)]);
    }
  #if HOR_ATOM == 0
  barrier(CLK_LOCAL_MEM_FENCE);
  #endif
  for(int i = 0; i < HOR_PAIRS_PER_THREAD; i++)
    {
      l_in_image[change_offset(l_in_image_pos + i * 2 + 4)] += DELTA * (l_in_image[change_offset(l_in_image_pos + i * 2 + 5)] + l_in_image[change_offset(l_in_image_pos + i * 2 + 3)]);
    }
  // save output coefficients
  for(int i = 0; i < HOR_PAIRS_PER_THREAD; i++)
    {
      out_image_l[i * out_stride] = ZETA1 * l_in_image[change_offset(l_in_image_pos + i * 2 + 4)];
      out_image_h[i * out_stride] = ZETA2 * l_in_image[change_offset(l_in_image_pos + i * 2 + 5)];
    }
}
#endif


#ifdef wavelet_hor_basic
__kernel void wavelet_hor(__global float *out_image, __global float *in_image, uint width, uint height, uint depth)
{
  size_t lx = get_local_id(0);
  size_t ly = get_local_id(1);
	

  #ifndef WAVELET_OUTPUT_INTERLACED
	  size_t gx = get_global_id(0);
	  size_t gy = get_global_id(1);
    int gx_l = gx * HOR_PAIRS_PER_THREAD;
    int l_to_h_stride = width>>(1+depth);
    int out_stride = 1;
    int block_stride = HOR_PAIRS_PER_GROUP;
	  int in_stride = 1;
  #else
	  size_t gx = get_global_id(0) << depth;
	  size_t gy = get_global_id(1) << depth;
    int gx_l = gx * HOR_PIXELS_PER_THREAD;
    int l_to_h_stride = 1 << depth;
    int out_stride = 1 << (1 + depth);
    int block_stride = HOR_PIXELS_PER_GROUP << depth;
	  int in_stride = 1 << depth;
  #endif

  __local float temp_image[HOR_LOCAL_SIZE_ALIGN];

  __global float *g_in_image = in_image + lx * in_stride + gy * width;
  __global float *g_out_image_l = out_image + gx_l + gy * width;
  __global float *g_out_image_h = g_out_image_l + l_to_h_stride;
  size_t l_load_image_pos = lx + ly * HOR_LOCAL_SIZE_X;
  size_t l_proc_image_pos = lx * HOR_PIXELS_PER_THREAD + ly * HOR_LOCAL_SIZE_X;
  if(((width >> depth) / HOR_PIXELS_PER_GROUP) == 1)
	{
	  load_hor_lr(g_in_image, temp_image, lx, l_load_image_pos, in_stride);
    #if HOR_ATOM == 0
	  barrier(CLK_LOCAL_MEM_FENCE);
    #endif
    #if HOR_PROC_TYPE == PROC_TYPE_LAAN
      proc_hor_start(temp_image, l_proc_image_pos, lx);
    #endif
	  proc_hor(temp_image, g_out_image_l, g_out_image_h, l_proc_image_pos, out_stride);
	  return;
	}
  load_hor_l(g_in_image, temp_image, lx, l_load_image_pos, in_stride);
  #if HOR_ATOM == 0
  barrier(CLK_LOCAL_MEM_FENCE);
  #endif
  #if HOR_PROC_TYPE == PROC_TYPE_LAAN
    proc_hor_start(temp_image, l_proc_image_pos, lx);
  #endif
  proc_hor(temp_image, g_out_image_l, g_out_image_h, l_proc_image_pos, out_stride);
  g_in_image += (HOR_PIXELS_PER_GROUP + 3) * in_stride;
  g_out_image_l += block_stride;
  g_out_image_h += block_stride;
  for(int i = 1; i < ((width >> depth) / HOR_PIXELS_PER_GROUP) - 1; i++)
  {
    #if HOR_ATOM == 0
    barrier(CLK_LOCAL_MEM_FENCE);
    #endif
    load_hor_m(g_in_image, temp_image, lx, l_load_image_pos, in_stride);
    #if HOR_ATOM == 0
    barrier(CLK_LOCAL_MEM_FENCE);
    #endif
    proc_hor(temp_image, g_out_image_l, g_out_image_h, l_proc_image_pos, out_stride);
    g_in_image += HOR_PIXELS_PER_GROUP * in_stride;
    g_out_image_l += block_stride;
    g_out_image_h += block_stride;
  }
  #if HOR_ATOM == 0
  barrier(CLK_LOCAL_MEM_FENCE);
  #endif
  load_hor_r(g_in_image, temp_image, lx, l_load_image_pos, in_stride);
  #if HOR_ATOM == 0
  barrier(CLK_LOCAL_MEM_FENCE);
  #endif
  proc_hor(temp_image, g_out_image_l, g_out_image_h, l_proc_image_pos, out_stride);
}
#endif


#ifdef wavelet_hor_split
__kernel void wavelet_hor(__global float *out_image, __global float *in_image, uint width, uint height, uint depth)
{
  size_t lx = get_local_id(0);
  size_t ly = get_local_id(1);
  size_t grx = get_group_id(0);
  
  #ifndef WAVELET_OUTPUT_INTERLACED
	size_t gx = get_global_id(0);
	size_t gy = get_global_id(1);
    int gx_l = gx * HOR_PAIRS_PER_THREAD;
    int l_to_h_stride = width>>(1+depth);
    int out_stride = 1;
	int in_stride = 1;
  #else
	size_t gx = get_global_id(0) << depth;
	size_t gy = get_global_id(1) << depth;
    int gx_l = gx * HOR_PIXELS_PER_THREAD;
    int l_to_h_stride = 1 << depth;
    int out_stride = 1 << (1 + depth);
	int in_stride = 1 << depth;
  #endif

  __local float temp_image[HOR_LOCAL_SIZE_ALIGN];

  __global float *g_in_image = in_image + lx * in_stride + gy * width;
  __global float *g_out_image_l = out_image + gx_l + gy * width;
  __global float *g_out_image_h = g_out_image_l + l_to_h_stride;
  size_t l_load_image_pos = lx + ly * HOR_LOCAL_SIZE_X;
  size_t l_proc_image_pos = lx * HOR_PIXELS_PER_THREAD + ly * HOR_LOCAL_SIZE_X;

  if(grx == 0)
	{
	  if(grx == get_num_groups(0) - 1)
		{
		  load_hor_lr(g_in_image, temp_image, lx, l_load_image_pos, in_stride);
		}
	  else
		{
		  load_hor_l(g_in_image, temp_image, lx, l_load_image_pos, in_stride);
		}
	  #if HOR_ATOM == 0
	  barrier(CLK_LOCAL_MEM_FENCE);
    #endif
    #if HOR_PROC_TYPE == PROC_TYPE_LAAN
      proc_hor_start(temp_image, l_proc_image_pos, lx);
    #endif
	  proc_hor(temp_image, g_out_image_l, g_out_image_h, l_proc_image_pos, out_stride);
	}
  else if(grx != get_num_groups(0) - 1)
	{
	  g_in_image += (HOR_PIXELS_PER_GROUP * grx - 4) * in_stride;
	  load_hor_m_split(g_in_image, temp_image, lx, l_load_image_pos, in_stride);
	  #if HOR_ATOM == 0
	  barrier(CLK_LOCAL_MEM_FENCE);
    #endif
    #if HOR_PROC_TYPE == PROC_TYPE_LAAN
      proc_hor_start(temp_image, l_proc_image_pos, lx);
    #endif
	  proc_hor(temp_image, g_out_image_l, g_out_image_h, l_proc_image_pos, out_stride);
	}
  else
	{
	  g_in_image += (HOR_PIXELS_PER_GROUP * grx - 4) * in_stride;
	  load_hor_r_split(g_in_image, temp_image, lx, l_load_image_pos, in_stride);
	  #if HOR_ATOM == 0
	  barrier(CLK_LOCAL_MEM_FENCE);
    #endif
    #if HOR_PROC_TYPE == PROC_TYPE_LAAN
      proc_hor_start(temp_image, l_proc_image_pos, lx);
    #endif
	  proc_hor(temp_image, g_out_image_l, g_out_image_h, l_proc_image_pos, out_stride);
	}
}
#endif


__kernel void conv_rgba_gray(__global float4 *out_image, __global uchar16 *in_image, uint width, uint height)
{
	size_t g_id = get_global_id(0) + get_global_size(0) * get_global_id(1);
	if(g_id < width * height)
		{
			uchar16 in_pixels = in_image[g_id];
      float4 out_pixels;
      out_pixels.x = 0.299f * in_pixels.s2 + 0.587f * in_pixels.s1 + 0.114f * in_pixels.s0;
			out_pixels.y = 0.299f * in_pixels.s6 + 0.587f * in_pixels.s5 + 0.114f * in_pixels.s4;
      out_pixels.z = 0.299f * in_pixels.sa + 0.587f * in_pixels.s9 + 0.114f * in_pixels.s8;
      out_pixels.w = 0.299f * in_pixels.se + 0.587f * in_pixels.sd + 0.114f * in_pixels.sc;
      out_image[g_id] = out_pixels;
		}
}

#ifndef WAVELET_OUTPUT_INTERLACED
__kernel void conv_gray_rgba(__global uchar4 *out_image, __global float *in_image, uint width, uint height, uint depth)
{
	size_t g_x = get_global_id(0);
	size_t g_y = get_global_id(1);
	
	if((g_x >= width) || (g_y >= height)) return;
	
	size_t l_width = width >> depth;
	size_t l_height = height >> depth;
	
	size_t image_id = g_x + width * g_y;
	
	float in_pixel = in_image[image_id];
	float temp_val = fabs(in_pixel);
	
	if((g_x >= l_width) || (g_y >= l_height))
		{
			temp_val = log2(temp_val+1)*20;
		}

	unsigned char output_val = convert_uchar_sat(temp_val);

	uchar4 output = (uchar4){output_val, output_val, output_val, 0};
	out_image[image_id] = output;
}
#else
__kernel void conv_gray_rgba(__global uchar4 *out_image, __global float *in_image, uint width, uint height, uint depth)
{
	size_t g_x = get_global_id(0);
	size_t g_y = get_global_id(1);
	
	if((g_x >= width) || (g_y >= height)) return;
	int act_depth = depth - 1;
	int shift_x, shift_y;
	for(int i = 0; i < depth; i++)
    {
			shift_x = (g_x >= (width >> (i + 1))) ? 1 : 0;
			shift_y = (g_y >= (height >> (i + 1))) ? 1 : 0;
			if(shift_x + shift_y > 0)
				{
					act_depth = i;
					break;
				}
    }

	int gx_in = ((g_x % (width >> (act_depth + 1))) << (act_depth + 1)) + (shift_x << act_depth);
	int gy_in = ((g_y % (height >> (act_depth + 1))) << (act_depth + 1)) + (shift_y << act_depth);
	size_t in_image_id = gx_in + width * gy_in;
	
	size_t l_width = width >> depth;
	size_t l_height = height >> depth;
	
	size_t out_image_id = g_x + width * g_y;
	
	float in_pixel = in_image[in_image_id];
	float temp_val = fabs(in_pixel);
	
	if((g_x >= l_width) || (g_y >= l_height))
		{
			temp_val = log2(temp_val+1)*20;
		}

	unsigned char output_val = convert_uchar_sat(temp_val);

	uchar4 output = (uchar4){output_val, output_val, output_val, 0};
	out_image[out_image_id] = output;
}
#endif




#ifndef COMB_BLOCK_SIZE
  #define COMB_BLOCK_SIZE 256
#endif

#define COMB_BLOCK_SIZE_RATED (COMB_BLOCK_SIZE + 1)

#ifndef COMB_BLOCK_SIZE_X
  #define COMB_BLOCK_SIZE_X 32 
#endif

#define COMB_BLOCK_SIZE_Y (COMB_BLOCK_SIZE / COMB_BLOCK_SIZE_X)

#ifndef COMB_PAIRS_PER_THREAD_X
  #define COMB_PAIRS_PER_THREAD_X 1
#endif

#ifndef COMB_PAIRS_PER_THREAD_Y
  #define COMB_PAIRS_PER_THREAD_Y 1
#endif

#define COMB_PIXELS_PER_THREAD_X (COMB_PAIRS_PER_THREAD_X * 2)

#define COMB_PIXELS_PER_THREAD_Y (COMB_PAIRS_PER_THREAD_Y * 2)

#define COMB_PAIRS_PER_GROUP_X (COMB_PAIRS_PER_THREAD_X * COMB_BLOCK_SIZE_X)

#define COMB_PAIRS_PER_GROUP_Y (COMB_PAIRS_PER_THREAD_Y * COMB_BLOCK_SIZE_Y)

#define COMB_PIXELS_PER_GROUP_X (COMB_PAIRS_PER_GROUP_X * 2)

#define COMB_PIXELS_PER_GROUP_Y (COMB_PAIRS_PER_GROUP_Y * 2)

#define COMB_LOCAL_SIZE_X (COMB_PIXELS_PER_GROUP_X + 7)

#define COMB_LOCAL_SIZE_Y (COMB_PIXELS_PER_GROUP_Y + 7)

#define COMB_LOCAL_SIZE (COMB_LOCAL_SIZE_X * COMB_LOCAL_SIZE_Y)

#ifndef COMB_HOR_CORNERS_PROC
#define COMB_HOR_CORNERS_PROC 0
#endif

#define COMB_HOR_LOW_STRIDE (COMB_PAIRS_PER_GROUP_X)
#define COMB_HOR_LOW_STRIDE_LOCAL (COMB_PAIRS_PER_GROUP_X + 4)

#if (WARP_SIZE >= COMB_BLOCK_SIZE_X) && ((WARP_SIZE % COMB_BLOCK_SIZE_X) == 0) && ATOMIC_WARP
#define COMB_HOR_ATOM 1
#else
#define COMB_HOR_ATOM 0
#endif

void load_comb_x(__global float *in_image, __local float *temp_image, size_t lx, size_t x_stride)
{
	__global float *in_image_first = in_image;
	__global float *in_image_last = in_image;
  if(get_group_id(0) == 0)
    {
			if(lx < 4) in_image_first = in_image_first + (8 - 2 * lx) * x_stride;
    }
  else if(get_group_id(0) == get_num_groups(0) - 1)
    {
			if(lx > 3) in_image_last = in_image_last + (6 - 2 * lx) * x_stride;
    }
	
	temp_image[0] = in_image_first[0];
	for(int i = 1; i < COMB_PIXELS_PER_THREAD_X; i++)
		{
			temp_image[i * COMB_BLOCK_SIZE_X] = in_image[i * COMB_BLOCK_SIZE_X * x_stride];
		}
	if(lx < 7)
		{
			temp_image[COMB_PIXELS_PER_GROUP_X] = in_image_last[COMB_PIXELS_PER_GROUP_X * x_stride];
		}
}

void load_comb_x_corner(__global float *in_image, __local float *l_in_image, __local float *l_in_image_copy,  bool is_copy, size_t lx, size_t x_stride)
{
	__global float *in_image_first = in_image;
	__global float *in_image_last = in_image;
	float act_data;
  if(get_group_id(0) == 0)
    {
			if(lx < 4) in_image_first = in_image_first + (8 - 2 * lx) * x_stride;
    }
  else if(get_group_id(0) == get_num_groups(0) - 1)
    {
			if(lx > 3) in_image_last = in_image_last + (6 - 2 * lx) * x_stride;
    }
	if(!is_copy)
		{
			l_in_image[0] = in_image_first[0];
			for(int i = 1; i < COMB_PIXELS_PER_THREAD_X; i++)
				{
					l_in_image[i * COMB_BLOCK_SIZE_X] = in_image[i * COMB_BLOCK_SIZE_X * x_stride];
				}
			if(lx < 7)
				{
					l_in_image[COMB_PIXELS_PER_GROUP_X] = in_image_last[COMB_PIXELS_PER_GROUP_X * x_stride];
				}
		}
	else
		{
			act_data = in_image_first[0];
			l_in_image[0] = act_data;
			l_in_image_copy[0] = act_data;
			for(int i = 1; i < COMB_PIXELS_PER_THREAD_X; i++)
				{
					act_data = in_image[i * COMB_BLOCK_SIZE_X * x_stride];
					l_in_image[i * COMB_BLOCK_SIZE_X] = act_data;
					l_in_image_copy[i * COMB_BLOCK_SIZE_X] = act_data;
				}
			if(lx < 7)
				{
					act_data = in_image_last[COMB_PIXELS_PER_GROUP_X * x_stride];
					l_in_image[COMB_PIXELS_PER_GROUP_X] = act_data;
					l_in_image_copy[COMB_PIXELS_PER_GROUP_X] = act_data;
				}
		}
}

void load_comb_t(__global float *in_image, __local float *temp_image, size_t lx, size_t ly, size_t x_stride, size_t y_stride)
{
  #if COMB_HOR_CORNERS_PROC == 1
  load_comb_x_corner(in_image + 4 * y_stride, temp_image + 4 * COMB_LOCAL_SIZE_X, temp_image + (4 - 2 * ly) * COMB_LOCAL_SIZE_X, (ly < 5) && (ly != 0), lx, x_stride);
  for(int j = 1; j < COMB_PIXELS_PER_THREAD_Y; j++)
  #else
  for(int j = 0; j < COMB_PIXELS_PER_THREAD_Y; j++)
  #endif
    {
      load_comb_x(in_image + (4 + j * COMB_BLOCK_SIZE_Y) * y_stride, temp_image + (4 + j * COMB_BLOCK_SIZE_Y) * COMB_LOCAL_SIZE_X, lx, x_stride);
    }
  if(ly < 3)
    {
      load_comb_x(in_image + (4 + COMB_PIXELS_PER_GROUP_Y) * y_stride, temp_image + (4 + COMB_PIXELS_PER_GROUP_Y) * COMB_LOCAL_SIZE_X, lx, x_stride);
    }
}

void load_comb_m(__global float *in_image, __local float *temp_image, size_t lx, size_t ly, size_t x_stride, size_t y_stride)
{
  for(int j = 0; j < COMB_PIXELS_PER_THREAD_Y; j++)
    {
      load_comb_x(in_image + j * COMB_BLOCK_SIZE_Y * y_stride, temp_image + j * COMB_BLOCK_SIZE_Y * COMB_LOCAL_SIZE_X, lx, x_stride);
    }
  if(ly < 7)
    {
      load_comb_x(in_image + COMB_PIXELS_PER_GROUP_Y * y_stride, temp_image + COMB_PIXELS_PER_GROUP_Y * COMB_LOCAL_SIZE_X, lx, x_stride);
    }
}

void load_comb_b(__global float *in_image, __local float *temp_image, size_t lx, size_t ly, size_t x_stride, size_t y_stride)
{
  for(int j = 0; j < COMB_PIXELS_PER_THREAD_Y; j++)
    {
      load_comb_x(in_image + j * COMB_BLOCK_SIZE_Y * y_stride, temp_image + j * COMB_BLOCK_SIZE_Y * COMB_LOCAL_SIZE_X, lx, x_stride);
    }
  if(ly < 4)
    {
      #if COMB_HOR_CORNERS_PROC == 1
      load_comb_x_corner(in_image + COMB_PIXELS_PER_GROUP_Y * y_stride, temp_image + COMB_PIXELS_PER_GROUP_Y * COMB_LOCAL_SIZE_X, temp_image + (COMB_PIXELS_PER_GROUP_Y - 2 * ly + 6) * COMB_LOCAL_SIZE_X, (ly < 3), lx, x_stride);
      #else
      load_comb_x(in_image + COMB_PIXELS_PER_GROUP_Y * y_stride, temp_image + COMB_PIXELS_PER_GROUP_Y * COMB_LOCAL_SIZE_X, lx, x_stride);
      #endif
    }
}

void load_comb_e(__global float *in_image, __local float *temp_image, size_t lx, size_t ly)
{
  for(int j = 0; j < COMB_PIXELS_PER_THREAD_Y; j++)
    {
      for(int i = 0; i < COMB_PIXELS_PER_THREAD_X; i++)
        {
          temp_image[i * COMB_BLOCK_SIZE_X + j * COMB_BLOCK_SIZE_Y * COMB_LOCAL_SIZE_X] = 0.0f;
        }
      if(lx < 7)
        {
          temp_image[COMB_PIXELS_PER_GROUP_X + j * COMB_BLOCK_SIZE_Y * COMB_LOCAL_SIZE_X] = 0.0f;
        }
    }
  if(ly < 7)
    {
      for(int i = 0; i < COMB_PIXELS_PER_THREAD_X; i++)
        {
          temp_image[i * COMB_BLOCK_SIZE_X + COMB_PIXELS_PER_GROUP_Y * COMB_LOCAL_SIZE_X] = 0.0f;
        }
      if(lx < 7)
        {
          temp_image[COMB_PIXELS_PER_GROUP_X + COMB_PIXELS_PER_GROUP_Y * COMB_LOCAL_SIZE_X] = 0.0f;
        }
    }
}

void load_comb_eh(__global float *in_image, __local float *temp_image, size_t lx, size_t ly)
{
  for(int j = 0; j < COMB_PIXELS_PER_THREAD_Y; j++)
    {
      for(int i = 0; i < COMB_PIXELS_PER_THREAD_X; i++)
        {
          temp_image[i * COMB_BLOCK_SIZE_X + j * COMB_BLOCK_SIZE_Y * COMB_LOCAL_SIZE_X] = 255.0f;
        }
      if(lx < 7)
        {
          temp_image[COMB_PIXELS_PER_GROUP_X + j * COMB_BLOCK_SIZE_Y * COMB_LOCAL_SIZE_X] = 255.0f;
        }
    }
  if(ly < 7)
    {
      for(int i = 0; i < COMB_PIXELS_PER_THREAD_X; i++)
        {
          temp_image[i * COMB_BLOCK_SIZE_X + COMB_PIXELS_PER_GROUP_Y * COMB_LOCAL_SIZE_X] = 255.0f;
        }
      if(lx < 7)
        {
          temp_image[COMB_PIXELS_PER_GROUP_X + COMB_PIXELS_PER_GROUP_Y * COMB_LOCAL_SIZE_X] = 255.0f;
        }
    }
}

void comb_hor_copy_local_to_private(__local float *l_in_image, __private float *p_in_image)
{
  for(int i = 0; i < COMB_PIXELS_PER_THREAD_X + 5; i++)
    {
      p_in_image[i] = l_in_image[i + 1];
    }
  p_in_image[0] += ALPHA * (p_in_image[1] + l_in_image[0]);
  p_in_image[COMB_PIXELS_PER_THREAD_X + 4] += ALPHA * (p_in_image[COMB_PIXELS_PER_THREAD_X + 3] + l_in_image[COMB_PIXELS_PER_THREAD_X + 6]);
}

void comb_hor_copy_private_to_local(__local float *l_in_image,__private float *private_image)
{
  // save output coefficients
  for(int i = 0; i < COMB_PAIRS_PER_THREAD_X; i++)
    {
      #ifndef WAVELET_OUTPUT_INTERLACED
      l_in_image[i] = ZETA1 * private_image[i * 2 + 3];
      l_in_image[i + COMB_HOR_LOW_STRIDE] = ZETA2 * private_image[i * 2 + 4];
      #else
      l_in_image[i * 2] = ZETA1 * private_image[i * 2 + 3];
      l_in_image[i * 2 + 1] = ZETA2 * private_image[i * 2 + 4];
      #endif
    }
}

void comb_hor_copy_private_to_local_corners(__local float *l_in_image, __local float *l_in_image_copy, bool is_copy, __private float *private_image)
{
  // save output coefficients
  for(int i = 0; i < COMB_PAIRS_PER_THREAD_X; i++)
    {
      float2 act_data = {ZETA1 * private_image[i * 2 + 3], ZETA2 * private_image[i * 2 + 4]};
      #ifndef WAVELET_OUTPUT_INTERLACED
      l_in_image[i] = act_data.x;
      l_in_image[i + COMB_HOR_LOW_STRIDE] = act_data.y;
      #else
      l_in_image[i * 2] = act_data.x;
      l_in_image[i * 2 + 1] = act_data.y;
      #endif
      if(is_copy)
        {
          #ifndef WAVELET_OUTPUT_INTERLACED
          l_in_image_copy[i] = act_data.x;
          l_in_image_copy[i + COMB_HOR_LOW_STRIDE] = act_data.y;
          #else
          l_in_image_copy[i * 2] = act_data.x;
          l_in_image_copy[i * 2 + 1] = act_data.y;
          #endif
        }
    }
}

void comb_hor_blazewicz(__private float *private_image)
{
  for(int i = 1; i < COMB_PAIRS_PER_THREAD_X + 2; i++)
    {
      private_image[i * 2] += ALPHA * (private_image[i * 2 - 1] + private_image[i * 2 + 1]);
    }
  for(int i = 0; i < COMB_PAIRS_PER_THREAD_X + 2; i++)
    {
      private_image[i * 2 + 1] += BETA * (private_image[i * 2] + private_image[i * 2 + 2]);
    }
  for(int i = 1; i < COMB_PAIRS_PER_THREAD_X + 2; i++)
    {
      private_image[i * 2] += GAMMA * (private_image[i * 2 - 1] + private_image[i * 2 + 1]);
    }
  for(int i = 1; i < COMB_PAIRS_PER_THREAD_X + 1; i++)
    {
      private_image[i * 2 + 1] += DELTA * (private_image[i * 2] + private_image[i * 2 + 2]);
    }
}

void proc_comb_hor_blazewicz(__local float *l_in_image_in, __local float *l_in_image_out, size_t ly)
{
  // create private memory
  __private float private_image[COMB_PIXELS_PER_THREAD_X + 5];
  for(int i = 0; i < COMB_PIXELS_PER_THREAD_Y; i++)
    {
      comb_hor_copy_local_to_private(l_in_image_in + i * COMB_BLOCK_SIZE_Y * COMB_LOCAL_SIZE_X, private_image);
      // compute wavelet transform
      comb_hor_blazewicz(private_image);
      barrier(CLK_LOCAL_MEM_FENCE);
      comb_hor_copy_private_to_local(l_in_image_out + i * COMB_BLOCK_SIZE_Y * COMB_LOCAL_SIZE_X, private_image);
    }
  if(ly < 7)
    {
      comb_hor_copy_local_to_private(l_in_image_in + COMB_PIXELS_PER_GROUP_Y * COMB_LOCAL_SIZE_X, private_image);      
      comb_hor_blazewicz(private_image);
    }
  barrier(CLK_LOCAL_MEM_FENCE);
  if(ly < 7)
    {
      comb_hor_copy_private_to_local(l_in_image_out + COMB_PIXELS_PER_GROUP_Y * COMB_LOCAL_SIZE_X, private_image);
    }
}

void proc_comb_hor_blazewicz_m(__local float *l_in_image_in, __local float *l_in_image_out, size_t ly)
{
  // create private memory
  __private float private_image[COMB_PIXELS_PER_THREAD_X + 5];
  for(int i = 0; i < COMB_PIXELS_PER_THREAD_Y; i++)
    {
      comb_hor_copy_local_to_private(l_in_image_in + i * COMB_BLOCK_SIZE_Y * COMB_LOCAL_SIZE_X, private_image);
      // compute wavelet transform
      comb_hor_blazewicz(private_image);
      #if COMB_HOR_ATOM == 0
      barrier(CLK_LOCAL_MEM_FENCE);
      #endif
      comb_hor_copy_private_to_local(l_in_image_out + i * COMB_BLOCK_SIZE_Y * COMB_LOCAL_SIZE_X, private_image);
    }
  if(ly < 7)
    {
      comb_hor_copy_local_to_private(l_in_image_in + COMB_PIXELS_PER_GROUP_Y * COMB_LOCAL_SIZE_X, private_image);      
      comb_hor_blazewicz(private_image);
    }
  #if COMB_HOR_ATOM == 0
  barrier(CLK_LOCAL_MEM_FENCE);
  #endif
  if(ly < 7)
    {
      comb_hor_copy_private_to_local(l_in_image_out + COMB_PIXELS_PER_GROUP_Y * COMB_LOCAL_SIZE_X, private_image);
    }
}

void proc_comb_hor_blazewicz_t(__local float *l_in_image_in, __local float *l_in_image_out, size_t ly)
{
  // create private memory
  __private float private_image[COMB_PIXELS_PER_THREAD_X + 5];
  comb_hor_copy_local_to_private(l_in_image_in + 4 * COMB_LOCAL_SIZE_X, private_image);
  // compute wavelet transform
  comb_hor_blazewicz(private_image);
  #if COMB_HOR_ATOM == 0
  barrier(CLK_LOCAL_MEM_FENCE);
  #endif
  comb_hor_copy_private_to_local_corners(l_in_image_out + 4 * COMB_LOCAL_SIZE_X, l_in_image_out + (4 - 2 * ly) * COMB_LOCAL_SIZE_X, ((ly < 5) && (ly != 0)), private_image);
  for(int i = 1; i < COMB_PIXELS_PER_THREAD_Y; i++)
    {
      comb_hor_copy_local_to_private(l_in_image_in + (4 + i * COMB_BLOCK_SIZE_Y) * COMB_LOCAL_SIZE_X, private_image);
      // compute wavelet transform
      comb_hor_blazewicz(private_image);
      #if COMB_HOR_ATOM == 0
      barrier(CLK_LOCAL_MEM_FENCE);
      #endif
      comb_hor_copy_private_to_local(l_in_image_out + (4 + i * COMB_BLOCK_SIZE_Y) * COMB_LOCAL_SIZE_X, private_image);
    }
  if(ly < 3)
    {
      comb_hor_copy_local_to_private(l_in_image_in + (4 + COMB_PIXELS_PER_GROUP_Y) * COMB_LOCAL_SIZE_X, private_image);      
      comb_hor_blazewicz(private_image);
    }
  #if COMB_HOR_ATOM == 0
  barrier(CLK_LOCAL_MEM_FENCE);
  #endif
  if(ly < 3)
    {
      comb_hor_copy_private_to_local(l_in_image_out + (4 + COMB_PIXELS_PER_GROUP_Y) * COMB_LOCAL_SIZE_X, private_image);
    }
}

void proc_comb_hor_blazewicz_b(__local float *l_in_image_in, __local float *l_in_image_out, size_t ly)
{
  // create private memory
  __private float private_image[COMB_PIXELS_PER_THREAD_X + 5];
  for(int i = 0; i < COMB_PIXELS_PER_THREAD_Y; i++)
    {
      comb_hor_copy_local_to_private(l_in_image_in + i * COMB_BLOCK_SIZE_Y * COMB_LOCAL_SIZE_X, private_image);
      // compute wavelet transform
      comb_hor_blazewicz(private_image);
      #if COMB_HOR_ATOM == 0
      barrier(CLK_LOCAL_MEM_FENCE);
      #endif
      comb_hor_copy_private_to_local(l_in_image_out + i * COMB_BLOCK_SIZE_Y * COMB_LOCAL_SIZE_X, private_image);
    }
  if(ly < 4)
    {
      comb_hor_copy_local_to_private(l_in_image_in + COMB_PIXELS_PER_GROUP_Y * COMB_LOCAL_SIZE_X, private_image);      
      comb_hor_blazewicz(private_image);
    }
  #if COMB_HOR_ATOM == 0
  barrier(CLK_LOCAL_MEM_FENCE);
  #endif
  if(ly < 4)
    {
      comb_hor_copy_private_to_local_corners(l_in_image_out + COMB_PIXELS_PER_GROUP_Y * COMB_LOCAL_SIZE_X, l_in_image_out + (COMB_PIXELS_PER_GROUP_Y - 2 * ly + 6) * COMB_LOCAL_SIZE_X, (ly < 3), private_image);
    }
}

void proc_comb_vert_blazewicz(__local float *local_image, __global float *out_image_h, __global float *out_image_l, size_t y_stride)
{
  // create private memory
  __private float private_image[COMB_PIXELS_PER_THREAD_Y + 5];
  for (int i = 0; i < COMB_PIXELS_PER_THREAD_Y + 5; i++)
    {
      private_image[i] = local_image[(i + 1) * COMB_LOCAL_SIZE_X];
    }
  // compute wavelet transform
  private_image[0] += ALPHA * (private_image[1] + local_image[0]);
  private_image[COMB_PIXELS_PER_THREAD_Y + 4] += ALPHA * (private_image[COMB_PIXELS_PER_THREAD_Y + 3] + local_image[(COMB_PIXELS_PER_THREAD_Y + 6) * COMB_LOCAL_SIZE_X]);
  for(int i = 1; i < COMB_PAIRS_PER_THREAD_Y + 2; i++)
    {
      private_image[i * 2] += ALPHA * (private_image[i * 2 - 1] + private_image[i * 2 + 1]);
    }
  for(int i = 0; i < COMB_PAIRS_PER_THREAD_Y + 2; i++)
    {
      private_image[i * 2 + 1] += BETA * (private_image[i * 2] + private_image[i * 2 + 2]);
    }
  for(int i = 1; i < COMB_PAIRS_PER_THREAD_Y + 2; i++)
    {
      private_image[i * 2] += GAMMA * (private_image[i * 2 - 1] + private_image[i * 2 + 1]);
    }
  for(int i = 1; i < COMB_PAIRS_PER_THREAD_Y + 1; i++)
    {
      private_image[i * 2 + 1] += DELTA * (private_image[i * 2] + private_image[i * 2 + 2]);
    }
  // save output coefficients
  for(int i = 0; i < COMB_PAIRS_PER_THREAD_Y; i++)
    {
      out_image_h[i * y_stride] = ZETA1 * private_image[i * 2 + 3];
      out_image_l[i * y_stride] = ZETA2 * private_image[i * 2 + 4];
    }
}



#define COMB_PIXELS_PER_THREAD_Y_RATED (COMB_PIXELS_PER_THREAD_Y * 2)
#define COMB_PAIRS_PER_THREAD_Y_RATED (COMB_PAIRS_PER_THREAD_Y * 2)

void proc_comb_vert_rated_blazewicz(__local float *local_image, __global float *out_image_h, __global float *out_image_l, uint stride)
{
  // create private memory
  __private float private_image[COMB_PIXELS_PER_THREAD_Y_RATED + 5];
  for(int i = 0; i < COMB_PIXELS_PER_THREAD_Y_RATED + 5; i++)
    {
      private_image[i] = local_image[(i + 1) * COMB_LOCAL_SIZE_X];
    }
  // compute wavelet transform
  private_image[0] += ALPHA * (private_image[1] + local_image[0]);
  private_image[COMB_PIXELS_PER_THREAD_Y_RATED + 4] += ALPHA * (private_image[COMB_PIXELS_PER_THREAD_Y_RATED + 3] + local_image[(COMB_PIXELS_PER_THREAD_Y_RATED + 6) * COMB_LOCAL_SIZE_X]);
  for(int i = 1; i < COMB_PAIRS_PER_THREAD_Y_RATED + 2; i++)
    {
      private_image[i * 2] += ALPHA * (private_image[i * 2 - 1] + private_image[i * 2 + 1]);
    }
  for(int i = 0; i < COMB_PAIRS_PER_THREAD_Y_RATED + 2; i++)
    {
      private_image[i * 2 + 1] += BETA * (private_image[i * 2] + private_image[i * 2 + 2]);
    }
  for(int i = 1; i < COMB_PAIRS_PER_THREAD_Y_RATED + 2; i++)
    {
      private_image[i * 2] += GAMMA * (private_image[i * 2 - 1] + private_image[i * 2 + 1]);
    }
  for(int i = 1; i < COMB_PAIRS_PER_THREAD_Y_RATED + 1; i++)
    {
      private_image[i * 2 + 1] += DELTA * (private_image[i * 2] + private_image[i * 2 + 2]);
    }
  // save output coefficients
  for(int i = 0; i < COMB_PAIRS_PER_THREAD_Y_RATED; i++)
    {
      out_image_h[i * stride] = ZETA1 * private_image[i * 2 + 3];
      out_image_l[i * stride] = ZETA2 * private_image[i * 2 + 4];
			/*out_image_h[i * width] = 0.0f;
      out_image_l[i * width] = 0.0f;*/
    }
}

#if defined(wavelet_sep_out_vert_rated) || defined(wavelet_sep_out_basic)
__kernel void wavelet_block(__global float *out_image, __global float *in_image, uint width, uint height, uint depth)
{
  size_t gx = get_global_id(0);
  size_t gy = get_global_id(1);
  size_t grx = get_group_id(0);
  size_t gry = get_group_id(1);
  size_t lx = get_local_id(0);
  size_t ly = get_local_id(1);

  __local float temp_image[COMB_LOCAL_SIZE];

  size_t max_width = width >> depth;
  size_t max_height = height >> depth;

  #ifndef WAVELET_OUTPUT_INTERLACED
	  size_t x_in_stride = 1;
	  size_t x_out_stride = 1;
	
	  size_t y_in_stride = width;
	  size_t y_out_stride = width;

	  size_t x_l_stride = 1;
	  size_t y_l_stride = 1;

	  int lrx_stride = width >> (1 + depth);
	  int lhy_stride = (height >> (1 + depth)) * width;
  #else
	  size_t x_in_stride = 1 << depth;
	  size_t x_out_stride = 2 * x_in_stride;
	
	  size_t y_in_stride = width * x_in_stride;
	  size_t y_out_stride = 2 * y_in_stride;

	  size_t x_l_stride = 2;
	  size_t y_l_stride = 2;
	
	  int lrx_stride = COMB_BLOCK_SIZE_X << depth;
    int lhy_stride = y_in_stride;
  #endif

  int grx_size = COMB_BLOCK_SIZE_X * 2;
  int gry_size = COMB_BLOCK_SIZE_Y * 2;
  int grx_start = grx_size * grx;
  int gry_start = gry_size * gry;
  int gx_in = grx_start + (lx - 4);
  int gy_in = gry_start + (ly - 4);

  int tx_out_stride = COMB_PAIRS_PER_THREAD_X * x_out_stride;
  int ty_out_stride = COMB_PAIRS_PER_THREAD_Y * y_out_stride;

  int grx_out_stride = COMB_PAIRS_PER_GROUP_X * x_out_stride;
  int gry_out_stride = COMB_PAIRS_PER_GROUP_Y * y_out_stride;

  int bx_out_stride = COMB_BLOCK_SIZE_X * x_out_stride;
  int by_out_stride = COMB_BLOCK_SIZE_Y * y_out_stride;

  int bx_l_stride = COMB_BLOCK_SIZE_X * x_l_stride;
  int by_l_stride = COMB_BLOCK_SIZE_Y * y_l_stride;
   
  __global float *g_in_image = in_image + gx_in * x_in_stride + gy_in * y_in_stride;
  __local float *l_load_image_pos = temp_image + lx + ly * COMB_LOCAL_SIZE_X;

  __local float *l_proc_hor_image_pos_in = temp_image + lx * COMB_PIXELS_PER_THREAD_X + ly * COMB_LOCAL_SIZE_X;
  __local float *l_proc_hor_image_pos_out = temp_image + lx * COMB_PAIRS_PER_THREAD_X * x_l_stride + ly * COMB_LOCAL_SIZE_X;

  #if defined(wavelet_sep_out_basic)
    __global float *g_out_image_h_1 = out_image + gy * ty_out_stride + grx * grx_out_stride + lx * x_in_stride;
    __global float *g_out_image_l_1 = g_out_image_h_1 + lhy_stride;
    __global float *g_out_image_h_2 = g_out_image_h_1 + lrx_stride;
    __global float *g_out_image_l_2 = g_out_image_l_1 + lrx_stride;
    __local float *l_proc_vert_image_pos = temp_image + lx + ly * COMB_PIXELS_PER_THREAD_Y * COMB_LOCAL_SIZE_X;
  #elif defined(wavelet_sep_out_vert_rated)
    size_t lx_vert = get_local_id(0) + get_local_size(0) * (get_local_id(1) & 0x01);
    size_t ly_vert = get_local_id(1) >> 1;

    int ty_out_rated_stride = COMB_PAIRS_PER_THREAD_Y_RATED * y_out_stride;

    __global float *g_out_image_h = out_image + gry * gry_out_stride + ly_vert * ty_out_rated_stride + grx * grx_out_stride + lx * x_in_stride;
    __global float *g_out_image_l = g_out_image_h + lhy_stride;

    if((get_local_id(1) & 0x01) == 1)
      {
        g_out_image_h += lrx_stride;
        g_out_image_l += lrx_stride;
      }

    __local float *l_proc_vert_image_pos = temp_image + lx_vert + ly_vert * COMB_PIXELS_PER_THREAD_Y_RATED * COMB_LOCAL_SIZE_X;
  #endif
  
  if(gry == 0)
    {
      load_comb_t(g_in_image, l_load_image_pos, lx, ly, x_in_stride, y_in_stride);
      #if COMB_HOR_CORNERS_PROC == 1
        barrier(CLK_LOCAL_MEM_FENCE);
        proc_comb_hor_blazewicz(l_proc_hor_image_pos_in, l_proc_hor_image_pos_out, ly);
      #else
        #if COMB_HOR_ATOM == 0
        barrier(CLK_LOCAL_MEM_FENCE);
        #endif
        proc_comb_hor_blazewicz_t(l_proc_hor_image_pos_in, l_proc_hor_image_pos_out, ly);
      #endif
    }
  else if(gry == get_num_groups(1) - 1)
    {
      load_comb_b(g_in_image, l_load_image_pos, lx, ly, x_in_stride, y_in_stride);
      #if COMB_HOR_CORNERS_PROC == 1
        barrier(CLK_LOCAL_MEM_FENCE);
        proc_comb_hor_blazewicz(l_proc_hor_image_pos_in, l_proc_hor_image_pos_out, ly);
      #else
        #if COMB_HOR_ATOM == 0
        barrier(CLK_LOCAL_MEM_FENCE);
        #endif
        proc_comb_hor_blazewicz_b(l_proc_hor_image_pos_in, l_proc_hor_image_pos_out, ly);
      #endif
    }
  else
    {
      load_comb_m(g_in_image, l_load_image_pos, lx, ly, x_in_stride, y_in_stride);
      #if COMB_HOR_ATOM == 0
      barrier(CLK_LOCAL_MEM_FENCE);
      #endif
      proc_comb_hor_blazewicz_m(l_proc_hor_image_pos_in, l_proc_hor_image_pos_out, ly);
    }

  barrier(CLK_LOCAL_MEM_FENCE);
  #if defined(wavelet_sep_out_basic)
    for(int i = 0; i < COMB_PAIRS_PER_THREAD_X; i++)
      {
        proc_comb_vert_blazewicz(l_proc_vert_image_pos + i * bx_l_stride, g_out_image_h_1 + i * bx_out_stride, g_out_image_l_1 + i * bx_out_stride, y_out_stride);
        proc_comb_vert_blazewicz(l_proc_vert_image_pos + i * bx_l_stride + COMB_HOR_LOW_STRIDE, g_out_image_h_2 + i * bx_out_stride, g_out_image_l_2 + i * bx_out_stride, y_out_stride);
      }
  #elif defined(wavelet_sep_out_vert_rated)
    proc_comb_vert_rated_blazewicz(l_proc_vert_image_pos, g_out_image_h, g_out_image_l, y_out_stride);
  #endif
}
#endif



#define BLOCK_IN_LOCAL_SIZE_BASIC (COMB_BLOCK_SIZE_X * 2 * COMB_BLOCK_SIZE_Y * 2)
#define BLOCK_IN_LOCAL_SIZE_RATED (COMB_BLOCK_SIZE_X * 2 * COMB_BLOCK_SIZE_Y * 3)

#define COMB_BLOCK_LOC_LL (COMB_BLOCK_SIZE * 0)
#define COMB_BLOCK_LOC_LH (COMB_BLOCK_SIZE * 1)
#define COMB_BLOCK_LOC_HL (COMB_BLOCK_SIZE * 2)
#define COMB_BLOCK_LOC_HH (COMB_BLOCK_SIZE * 3)
#define COMB_BLOCK_LOC_LH2 (COMB_BLOCK_SIZE * 4)
#define COMB_BLOCK_LOC_HL2 (COMB_BLOCK_SIZE * 5)

#define COMB_BLOCK_LOC_T (-COMB_BLOCK_SIZE_X)
#define COMB_BLOCK_LOC_B (COMB_BLOCK_SIZE_X)
#define COMB_BLOCK_LOC_L (-1)
#define COMB_BLOCK_LOC_R (1)
#define COMB_BLOCK_LOC_TL (COMB_BLOCK_LOC_T+COMB_BLOCK_LOC_L)
#define COMB_BLOCK_LOC_BL (COMB_BLOCK_LOC_B+COMB_BLOCK_LOC_L)
#define COMB_BLOCK_LOC_TR (COMB_BLOCK_LOC_T+COMB_BLOCK_LOC_R)
#define COMB_BLOCK_LOC_BR (COMB_BLOCK_LOC_B+COMB_BLOCK_LOC_R)


float2 load_block_x(__global float *in_image, __local float *l_in_image, int gx, size_t x_stride, size_t max_width)
{
  __global float *in_image_l = in_image;
  __global float *in_image_h = in_image + x_stride;
  if(get_group_id(0) == 0)
    {
      if(gx<0)
        {
          in_image_l = in_image_l + (- 4 * gx) * x_stride;
          in_image_h = in_image_h + (-2 - 4 * gx) * x_stride;
        }
    }
  else if(get_group_id(0) == get_num_groups(0) - 1)
    {
			if(gx >= max_width)
        {
          in_image_l = in_image_l + (2 - 4 * (gx - (max_width - 1))) * x_stride;
          in_image_h = in_image_h + (-4 * (gx - (max_width - 1))) * x_stride;
        }
    }
  float2 act_data = (float2)(in_image_l[0],in_image_h[0]);
	l_in_image[COMB_BLOCK_LOC_LL] = act_data.x;
	l_in_image[COMB_BLOCK_LOC_LH] = act_data.y;
  return act_data;
}

float2 load_block_x_corner(__global float *in_image, __local float *l_in_image, __local float *l_in_image_copy, bool is_copy, int gx, size_t x_stride, size_t max_width)
{
  __global float *in_image_l = in_image;
  __global float *in_image_h = in_image + x_stride;
  if(get_group_id(0) == 0)
    {
			if(gx<0)
        {
          in_image_l = in_image_l + (- 4 * gx) * x_stride;
          in_image_h = in_image_h + (-2 - 4 * gx) * x_stride;
        }
    }
  else if(get_group_id(0) == get_num_groups(0) - 1)
    {
			if(gx >= max_width)
        {
          in_image_l = in_image_l + (2 - 4 * (gx - (max_width - 1))) * x_stride;
          in_image_h = in_image_h + (-4 * (gx - (max_width - 1))) * x_stride;
        }
    }
	float2 act_data = (float2)(in_image_l[0],in_image_h[0]);
	l_in_image[COMB_BLOCK_LOC_LL] = act_data.x;
	l_in_image[COMB_BLOCK_LOC_LH] = act_data.y;
	if(is_copy)
		{
			l_in_image_copy[COMB_BLOCK_LOC_LL] = act_data.x;
			l_in_image_copy[COMB_BLOCK_LOC_LH] = act_data.y;
		}
  return act_data;
}

/*void load_block_e(__global float *in_image, __local float *temp_image)
{
  temp_image[COMB_BLOCK_LOC_LL] = 0.0f;
	temp_image[COMB_BLOCK_LOC_LH] = 0.0f;
	temp_image[COMB_BLOCK_LOC_HL] = 0.0f;
	temp_image[COMB_BLOCK_LOC_HH] = 0.0f;
}*/

float4 load_block(__global float *in_image, __local float *temp_image, int gx, int gy, size_t x_stride, size_t y_stride, size_t max_width, size_t max_height)
{

  float4 out_data;
  if(get_group_id(1) == 0)
    {
      if(gy >= 0)
        {
          #if COMB_HOR_CORNERS_PROC == 1
          out_data = (float4)(load_block_x_corner(in_image, temp_image, temp_image + (-2 * gy) * COMB_BLOCK_SIZE_X, (gy > 0) && (gy <= 2), gx, x_stride, max_width),
                              load_block_x_corner(in_image + y_stride, temp_image + COMB_BLOCK_SIZE * 2, temp_image + (-1 - 2 * gy) * COMB_BLOCK_SIZE_X + COMB_BLOCK_SIZE * 2, (gy <= 1), gx, x_stride, max_width));
          #else
          out_data = (float4)(load_block_x(in_image, temp_image, gx, x_stride, max_width),
	                            load_block_x(in_image + y_stride, temp_image + COMB_BLOCK_SIZE * 2, gx, x_stride, max_width));
          #endif
        }
    }
  else if(get_group_id(1) == get_num_groups(1) - 1)
    {
      if(gy < max_height)
        {
          #if COMB_HOR_CORNERS_PROC == 1
          out_data = (float4)(load_block_x_corner(in_image, temp_image, temp_image + (1 + 2 * (max_height - 1 - gy)) * COMB_BLOCK_SIZE_X, (gy >= max_height - 2), gx, x_stride, max_width),
                              load_block_x_corner(in_image + y_stride, temp_image + COMB_BLOCK_SIZE * 2, temp_image + 2 * (max_height - 1 - gy) * COMB_BLOCK_SIZE_X + COMB_BLOCK_SIZE * 2, (gy < max_height - 1) && (gy >= max_height - 3), gx, x_stride, max_width));
          #else
          out_data = (float4)(load_block_x(in_image, temp_image, gx, x_stride, max_width),
	                            load_block_x(in_image + y_stride, temp_image + COMB_BLOCK_SIZE * 2, gx, x_stride, max_width));
          #endif
        }
    }
  else
    {
      out_data = (float4)(load_block_x(in_image, temp_image, gx, x_stride, max_width),
	                        load_block_x(in_image + y_stride, temp_image + COMB_BLOCK_SIZE * 2, gx, x_stride, max_width));
    }
  return out_data;
}



float4 proc_hor_laan(__local float *temp_image, float4 act_data)
{
  act_data.y += ALPHA*(act_data.x + temp_image[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R]);
  act_data.w += ALPHA*(act_data.z + temp_image[COMB_BLOCK_LOC_HL + COMB_BLOCK_LOC_R]);
  temp_image[COMB_BLOCK_LOC_LH] = act_data.y;
  temp_image[COMB_BLOCK_LOC_HH] = act_data.w;
  #if COMB_HOR_ATOM == 0
  barrier(CLK_LOCAL_MEM_FENCE);
  #endif
  act_data.x += BETA*(act_data.y + temp_image[COMB_BLOCK_LOC_LH + COMB_BLOCK_LOC_L]);
  act_data.z += BETA*(act_data.w + temp_image[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L]);
  temp_image[COMB_BLOCK_LOC_LL] = act_data.x;
  temp_image[COMB_BLOCK_LOC_HL] = act_data.z;
  #if COMB_HOR_ATOM == 0
  barrier(CLK_LOCAL_MEM_FENCE);
  #endif
  act_data.y += GAMMA*(act_data.x + temp_image[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R]);
  act_data.w += GAMMA*(act_data.z + temp_image[COMB_BLOCK_LOC_HL + COMB_BLOCK_LOC_R]);
  temp_image[COMB_BLOCK_LOC_LH] = act_data.y;
  temp_image[COMB_BLOCK_LOC_HH] = act_data.w;
  #if COMB_HOR_ATOM == 0
  barrier(CLK_LOCAL_MEM_FENCE);
  #endif
  act_data.x += DELTA*(act_data.y + temp_image[COMB_BLOCK_LOC_LH + COMB_BLOCK_LOC_L]);
  act_data.z += DELTA*(act_data.w + temp_image[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L]);
  temp_image[COMB_BLOCK_LOC_LL] = act_data.x;
  temp_image[COMB_BLOCK_LOC_HL] = act_data.z;
  barrier(CLK_LOCAL_MEM_FENCE);
  return act_data;
}

float4 proc_vert_laan(__local float *temp_image, __local float *l_t_image, __local float *l_b_image, float4 act_data)
{
  act_data.z += ALPHA*(act_data.x + l_b_image[COMB_BLOCK_LOC_LL]);
  act_data.w += ALPHA*(act_data.y + l_b_image[COMB_BLOCK_LOC_LH]);
  temp_image[COMB_BLOCK_LOC_HL] = act_data.z;
  temp_image[COMB_BLOCK_LOC_HH] = act_data.w;
  barrier(CLK_LOCAL_MEM_FENCE);
  act_data.x += BETA*(act_data.z + l_t_image[COMB_BLOCK_LOC_HL]);
  act_data.y += BETA*(act_data.w + l_t_image[COMB_BLOCK_LOC_HH]);
  temp_image[COMB_BLOCK_LOC_LL] = act_data.x;
  temp_image[COMB_BLOCK_LOC_LH] = act_data.y;
  barrier(CLK_LOCAL_MEM_FENCE);
  act_data.z += GAMMA*(act_data.x + l_b_image[COMB_BLOCK_LOC_LL]);
  act_data.w += GAMMA*(act_data.y + l_b_image[COMB_BLOCK_LOC_LH]);
  temp_image[COMB_BLOCK_LOC_HL] = act_data.z;
  temp_image[COMB_BLOCK_LOC_HH] = act_data.w;
  barrier(CLK_LOCAL_MEM_FENCE);
  act_data.x += DELTA*(act_data.z + l_t_image[COMB_BLOCK_LOC_HL]);
  act_data.y += DELTA*(act_data.w + l_t_image[COMB_BLOCK_LOC_HH]);
  return act_data;
}

#ifdef wavelet_block_in_3step_24mac
#define BLOCK_IN_LOCAL_SIZE BLOCK_IN_LOCAL_SIZE_BASIC
float4 proc_block(__local float *temp_image, __local float *temp_image_t, __local float *temp_image_b, float4 act_data)
{
  act_data.w += ALPHA*(act_data.y + act_data.z + temp_image_b[COMB_BLOCK_LOC_LH] + temp_image[COMB_BLOCK_LOC_HL + COMB_BLOCK_LOC_R]) + 
                ALPHA*ALPHA*(act_data.x + temp_image[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R] + temp_image_b[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R] + temp_image_b[COMB_BLOCK_LOC_LL]);
  temp_image[COMB_BLOCK_LOC_HH] = act_data.w;
  
  barrier(CLK_LOCAL_MEM_FENCE);
  
  act_data.y += ALPHA*(act_data.x + temp_image[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R]) + 
                BETA*(act_data.w + temp_image_t[COMB_BLOCK_LOC_HH]);
  act_data.z += ALPHA*(act_data.x + temp_image_b[COMB_BLOCK_LOC_LL]) + 
                BETA*(act_data.w + temp_image[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L]);
  temp_image[COMB_BLOCK_LOC_LH] = act_data.y;
  temp_image[COMB_BLOCK_LOC_HL] = act_data.z;
  
  barrier(CLK_LOCAL_MEM_FENCE);
  
  act_data.x += BETA*(act_data.y + act_data.z + temp_image[COMB_BLOCK_LOC_LH + COMB_BLOCK_LOC_L] + temp_image_t[COMB_BLOCK_LOC_HL]) - 
                BETA*BETA*(act_data.w + temp_image_t[COMB_BLOCK_LOC_HH] + temp_image_t[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L] + temp_image[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L]);
  temp_image[COMB_BLOCK_LOC_LL] = act_data.x;
  
  barrier(CLK_LOCAL_MEM_FENCE);

  act_data.w += GAMMA*(act_data.y + act_data.z + temp_image_b[COMB_BLOCK_LOC_LH] + temp_image[COMB_BLOCK_LOC_HL + COMB_BLOCK_LOC_R]) + 
                GAMMA*GAMMA*(act_data.x + temp_image[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R] + temp_image_b[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R] + temp_image_b[COMB_BLOCK_LOC_LL]);
  temp_image[COMB_BLOCK_LOC_HH] = act_data.w;
  
  barrier(CLK_LOCAL_MEM_FENCE);
  
  act_data.y += GAMMA*(act_data.x + temp_image[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R]) + 
                DELTA*(act_data.w + temp_image_t[COMB_BLOCK_LOC_HH]);
  act_data.z += GAMMA*(act_data.x + temp_image_b[COMB_BLOCK_LOC_LL]) + 
                DELTA*(act_data.w + temp_image[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L]);
  temp_image[COMB_BLOCK_LOC_LH] = act_data.y;
  temp_image[COMB_BLOCK_LOC_HL] = act_data.z;
  
  barrier(CLK_LOCAL_MEM_FENCE);
  
  act_data.x += DELTA*(act_data.y + act_data.z + temp_image[COMB_BLOCK_LOC_LH + COMB_BLOCK_LOC_L] + temp_image_t[COMB_BLOCK_LOC_HL]) - 
                DELTA*DELTA*(act_data.w + temp_image_t[COMB_BLOCK_LOC_HH] + temp_image_t[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L] + temp_image[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L]);
  return act_data;
}
#endif

#ifdef wavelet_block_in_2step_24mac
#define BLOCK_IN_LOCAL_SIZE BLOCK_IN_LOCAL_SIZE_RATED
float4 proc_block(__local float *temp_image, __local float *temp_image_t, __local float *temp_image_b, float4 act_data)
{
  act_data.w += ALPHA*(act_data.y + act_data.z + temp_image_b[COMB_BLOCK_LOC_LH] + temp_image[COMB_BLOCK_LOC_HL + COMB_BLOCK_LOC_R]) + 
                ALPHA*ALPHA*(act_data.x + temp_image[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R] + temp_image_b[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R] + temp_image_b[COMB_BLOCK_LOC_LL]);
  act_data.y += ALPHA*(act_data.x + temp_image[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R]);
  act_data.z += ALPHA*(act_data.x + temp_image_b[COMB_BLOCK_LOC_LL]);
  temp_image[COMB_BLOCK_LOC_LH2] = act_data.y;
  temp_image[COMB_BLOCK_LOC_HL2] = act_data.z;
  temp_image[COMB_BLOCK_LOC_HH] = act_data.w;
  barrier(CLK_LOCAL_MEM_FENCE);
  act_data.x += BETA*(act_data.y + act_data.z + temp_image[COMB_BLOCK_LOC_LH2 + COMB_BLOCK_LOC_L] + temp_image_t[COMB_BLOCK_LOC_HL2]) + 
                BETA*BETA*(act_data.w + temp_image_t[COMB_BLOCK_LOC_HH] + temp_image_t[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L] + temp_image[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L]);
  act_data.y += BETA*(act_data.w + temp_image_t[COMB_BLOCK_LOC_HH]);
  act_data.z += BETA*(act_data.w + temp_image[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L]);
  temp_image[COMB_BLOCK_LOC_LL] = act_data.x;
  temp_image[COMB_BLOCK_LOC_LH] = act_data.y;
  temp_image[COMB_BLOCK_LOC_HL] = act_data.z;
  barrier(CLK_LOCAL_MEM_FENCE);

  act_data.w += GAMMA*(act_data.y + act_data.z + temp_image_b[COMB_BLOCK_LOC_LH] + temp_image[COMB_BLOCK_LOC_HL + COMB_BLOCK_LOC_R]) + 
                GAMMA*GAMMA*(act_data.x + temp_image[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R] + temp_image_b[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R] + temp_image_b[COMB_BLOCK_LOC_LL]);
  act_data.y += GAMMA*(act_data.x + temp_image[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R]);
  act_data.z += GAMMA*(act_data.x + temp_image_b[COMB_BLOCK_LOC_LL]);
  temp_image[COMB_BLOCK_LOC_LH2] = act_data.y;
  temp_image[COMB_BLOCK_LOC_HL2] = act_data.z;
  temp_image[COMB_BLOCK_LOC_HH] = act_data.w;
  barrier(CLK_LOCAL_MEM_FENCE);
  
  act_data.x += DELTA*(act_data.y + act_data.z + temp_image[COMB_BLOCK_LOC_LH2 + COMB_BLOCK_LOC_L] + temp_image_t[COMB_BLOCK_LOC_HL2]) + 
                DELTA*DELTA*(act_data.w + temp_image_t[COMB_BLOCK_LOC_HH] + temp_image_t[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L] + temp_image[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L]);
  act_data.y += DELTA*(act_data.w + temp_image_t[COMB_BLOCK_LOC_HH]);
  act_data.z += DELTA*(act_data.w + temp_image[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L]);
  
  return act_data;
}
#endif

#ifdef wavelet_block_in_2step_20mac
#define BLOCK_IN_LOCAL_SIZE BLOCK_IN_LOCAL_SIZE_RATED
float4 proc_block(__local float *temp_image, __local float *temp_image_t, __local float *temp_image_b, float4 act_data)
{
  act_data.y += ALPHA*(act_data.x + temp_image[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R]);
  act_data.w += ALPHA*(act_data.y + act_data.z + temp_image_b[COMB_BLOCK_LOC_LH] + temp_image[COMB_BLOCK_LOC_HL + COMB_BLOCK_LOC_R]) + 
                ALPHA*ALPHA*(temp_image_b[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R] + temp_image_b[COMB_BLOCK_LOC_LL]);
  act_data.z += ALPHA*(act_data.x + temp_image_b[COMB_BLOCK_LOC_LL]);
  temp_image[COMB_BLOCK_LOC_LH2] = act_data.y;
  temp_image[COMB_BLOCK_LOC_HL2] = act_data.z;
  temp_image[COMB_BLOCK_LOC_HH] = act_data.w;
  barrier(CLK_LOCAL_MEM_FENCE);
  act_data.y += BETA*(act_data.w + temp_image_t[COMB_BLOCK_LOC_HH]);
  act_data.x += BETA*(act_data.y + act_data.z + temp_image[COMB_BLOCK_LOC_LH2 + COMB_BLOCK_LOC_L] + temp_image_t[COMB_BLOCK_LOC_HL2]) + 
                BETA*BETA*(temp_image[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L] + temp_image[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L]);
  act_data.z += BETA*(act_data.w + temp_image[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L]);
  temp_image[COMB_BLOCK_LOC_LL] = act_data.x;
  temp_image[COMB_BLOCK_LOC_LH] = act_data.y;
  temp_image[COMB_BLOCK_LOC_HL] = act_data.z;
  barrier(CLK_LOCAL_MEM_FENCE);
  act_data.y += GAMMA*(act_data.x + temp_image[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R]);
  act_data.w += GAMMA*(act_data.y + act_data.z + temp_image_b[COMB_BLOCK_LOC_LH] + temp_image[COMB_BLOCK_LOC_HL + COMB_BLOCK_LOC_R]) + 
                GAMMA*GAMMA*(temp_image_b[COMB_BLOCK_LOC_LL + COMB_BLOCK_LOC_R] + temp_image_b[COMB_BLOCK_LOC_LL]);
  act_data.z += GAMMA*(act_data.x + temp_image_b[COMB_BLOCK_LOC_LL]);
  temp_image[COMB_BLOCK_LOC_LH2] = act_data.y;
  temp_image[COMB_BLOCK_LOC_HL2] = act_data.z;
  temp_image[COMB_BLOCK_LOC_HH] = act_data.w;
  barrier(CLK_LOCAL_MEM_FENCE);
  act_data.y += DELTA*(act_data.w + temp_image_t[COMB_BLOCK_LOC_HH]);
  act_data.x += DELTA*(act_data.y + act_data.z + temp_image[COMB_BLOCK_LOC_LH2 + COMB_BLOCK_LOC_L] + temp_image_t[COMB_BLOCK_LOC_HL2]) + 
                DELTA*DELTA*(temp_image_t[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L] + temp_image[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L]);
  act_data.z += DELTA*(act_data.w + temp_image[COMB_BLOCK_LOC_HH + COMB_BLOCK_LOC_L]);
  
  return act_data;
}
#endif

#ifdef wavelet_sep_in_basic
#define BLOCK_IN_LOCAL_SIZE BLOCK_IN_LOCAL_SIZE_BASIC
float4 proc_block(__local float *temp_image, __local float *temp_image_t, __local float *temp_image_b, float4 act_data)
  {
    float4 out_data = proc_hor_laan(temp_image, act_data);
    out_data = proc_vert_laan(temp_image, temp_image_t, temp_image_b, out_data);
    return out_data;
  }
#endif

#if defined(wavelet_block_in_3step_24mac) || defined(wavelet_sep_in_basic) || defined(wavelet_block_in_2step_24mac) || defined(wavelet_block_in_2step_20mac)
__kernel void wavelet_block(__global float *out_image, __global float *in_image, uint width, uint height, uint depth)
{
  size_t grx = get_group_id(0);
  size_t gry = get_group_id(1);
  size_t lx = get_local_id(0);
  size_t ly = get_local_id(1);

  __local float temp_image[BLOCK_IN_LOCAL_SIZE];

  #ifndef WAVELET_OUTPUT_INTERLACED
	  size_t x_stride = 1;
	  size_t x_out_stride = 1;
	
	  size_t y_stride = width;
	  size_t y_out_stride = width;

	  int lr_stride = width >> (1 + depth);
	  int tb_stride = (height >> (1 + depth)) * width;
  #else
	  size_t x_stride = 1 << depth;
	  size_t x_out_stride = 2 * x_stride;
	
	  size_t y_stride = width * x_stride;
	  size_t y_out_stride = 2 * y_stride;
	
	  int lr_stride = x_stride;
    int tb_stride = y_stride;
  #endif

  size_t max_width = width >> (depth+1);
  size_t max_height = height >> (depth+1);

  int grx_size = (COMB_BLOCK_SIZE_X - 4);
  int gry_size = (COMB_BLOCK_SIZE_Y - 4);
  int grx_start = grx_size * grx;
  int gry_start = gry_size * gry;
  int gx = grx_start + (lx - 2);
  int gy = gry_start + (ly - 2);

  int x_in_stride = x_stride * 2;
  int y_in_stride = y_stride * 2;

  __global float *g_out_image_ll = out_image + gx * x_out_stride + gy * y_out_stride;
  __global float *g_out_image_lh = g_out_image_ll + lr_stride;
  __global float *g_out_image_hl = g_out_image_ll + tb_stride;
  __global float *g_out_image_hh = g_out_image_lh + tb_stride;

  size_t lxy = get_local_size(0) * get_local_id(1) + get_local_id(0);

  __global float *g_in_image = in_image + gx * x_in_stride + gy * y_in_stride;
  
  __local float *l_image_pos = temp_image + lxy;
  __local float *l_image_load_pos = temp_image + lxy;

  float4 out_data = load_block(g_in_image, l_image_load_pos, gx, gy, x_stride, y_stride, max_width, max_height);

  barrier(CLK_LOCAL_MEM_FENCE);
  __local float *l_t_image_pos = l_image_pos + COMB_BLOCK_LOC_T;
  __local float *l_b_image_pos = l_image_pos + COMB_BLOCK_LOC_B;
  #if COMB_HOR_CORNERS_PROC == 1
  if((gy < 0) || (gy >= max_height))
    {
      out_data = (float4){l_image_pos[COMB_BLOCK_LOC_LL], l_image_pos[COMB_BLOCK_LOC_LH], l_image_pos[COMB_BLOCK_LOC_HL], l_image_pos[COMB_BLOCK_LOC_HH]};
    }
  #else
  if(gy == 0)
    {
      l_t_image_pos = l_image_pos;
    }
  if(gy == max_height - 1)
    {
      l_b_image_pos = l_image_pos;
    }
  #endif
  out_data = proc_block(l_image_pos, l_t_image_pos, l_b_image_pos, out_data);

  if((lx >= 2) && (lx < COMB_BLOCK_SIZE_X - 2) && (ly >= 2) && (ly < COMB_BLOCK_SIZE_Y - 2) && (gx < max_width) && (gy < max_height))
    {
      g_out_image_ll[0] = out_data.x * ZETA1 * ZETA1;
      g_out_image_lh[0] = out_data.y;
      g_out_image_hl[0] = out_data.z;
      g_out_image_hh[0] = out_data.w * ZETA2 * ZETA2;
    }
}
#endif