#include "MainConfig.h"

MainConfig::MainConfig()
  {
    this->in_show = false;
    this->out_show = false;
    //this->device_type = DEV_TYPE_ALL;
    this->print_devices = false;
    this->kernel_id = DET_UNDEF;
    this->program_id = DET_UNDEF;
    this->frame_count = DET_UNDEF;
    this->cam_use = false;
    this->bench_type = CONF_BENCH_TYPE_NONE;
    this->engine = ENGINE_TYPE_OPENCL_COMB;
    const opt opts[] = {{(char *)"show-input", 0, false, (char *)"Show input video file while converting"},
                         {(char *)"show-output", 0, false, (char *)"Show output video file while converting"},
                         {(char *)"help", 'h', false, (char *)"Print this help"},
                         {(char *)"input", 'i', true, (char *)"Set path to input video file"},
                         {(char *)"device", 'd', true, (char *)"Set device {c,cpu,g,gpu,force_cpu,force_gpu}"},
                         {(char *)"print-devices", 0, false, (char *)"Print devices info"},
                         {(char *)"kernel", 'k', true, (char *)"Choose kernel_id in program"},
                         {(char *)"program", 'p', true, (char *)"Choose program to run {gpu,g,cpu,c}"},
                         {(char *)"frame-count", 0, true, (char *)"Set frame process count"},
                         {(char *)"cam", 0, true, (char *)"Camera number"},
                         {(char *)"hor-pairs", 0, true, (char *)"Pixel pairs to process by thread."},
                         {(char *)"vert-pairs", 0, true, (char *)"Pixel pairs to process by thread."},
                         {(char *)"hor-x", 0, true, (char *)"Width of group in horizontal transform."},
                         {(char *)"hor-y", 0, true, (char *)"Hieght of group in horizontal transform."},
                         {(char *)"vert-x", 0, true, (char *)"Width of group in vertical transform."},
                         {(char *)"vert-y", 0, true, (char *)"Hieght of group in vertical transform."},
                         {(char *)"comb-x", 0, true, (char *)"Width of group in combined transform."},
                         {(char *)"comb-y", 0, true, (char *)"Hieght of group in combined transform."},
                         {(char *)"hor-kernel", 0, true, (char *)"Horizontal pass kernel name."},
                         {(char *)"vert-kernel", 0, true, (char *)"Vertical pass kernel name."},
                         {(char *)"comb-kernel", 0, true, (char *)"Combined pass kernel named."},
                         {(char *)"benchmark", 0, true, (char *)"Set benchmark type {none,n,csv,c}"},
                         {(char *)"comb-corn-proc", 0, false, (char *)"Process horizontal corners."},
                         {(char *)"warp-size", 0, true, (char *)"Size of warp."},
                         {(char *)"non-atomic-warp", 0, false, (char *)"Warp instructions is not executed atomicaly."},
                         {(char *)"hor-proc", 0, true, (char *)"Set type of hor process {blaz_normal,bn,blaz_register,br,laan,l}"},
                         {(char *)"vert-proc", 0, true, (char *)"Set type of vert process {blaz_normal,bn,blaz_register,br,laan,l}"},
                         {(char *)"engine-type", 0, true, (char *)"Set benchmark type {opencl,openmp,openclcomb}"},
                         {(char *)"hor-func", 0, true, (char *)"Horizontal pass function name"},
                         {(char *)"vert-func", 0, true, (char *)"Vertical pass function name"},
                         {(char *)"interlaced", 0, false, (char *)"Wavelet output interlaced."},
						 {(char *)"max-depth", 0, true, (char *)"Maximum depth of transformation." }};
		this->parser.addOpts(opts, 32);
  }

MainConfig::~MainConfig()
  {
  }

void MainConfig::printHelp(const char *exec_name)
  {
    std::string header = std::string() + "\nUsage: " + exec_name + " -i=input_video_file [parameters]\n";
    this->parser.printHelp(header);
  }

bool MainConfig::isValid(std::stringstream *error)
  {
		bool result = true;
		return result;
  }

void MainConfig::loadFromArguments(int argc, char *argv[], std::stringstream *error)
  {
		std::string opt_value;
		int opt_num;
		int type_count = 0;
    int ival;
		for(int i = 1; i < argc; i++)
			{
				opt_num = parser.getArgOpt(argv[i], opt_value);
				switch(opt_num)
					{
						case 0:
							this->in_show = true;
						break;
						case 1:
							this->out_show = true;
						break;
						case 2:
							this->printHelp(argv[0]);
              exit(0);
						break;
						case 3:
              if(opt_value.length() == 0)
                {
                  if(error != NULL) *error << "Warning: Parameter \"input\" is empty.\n";
                }
              else
                {
							    this->in_file = opt_value;
                  this->cam_use = false;
                }
						break;
						case 4:
              if((opt_value.compare("gpu") == 0) || (opt_value.compare("g") == 0))
                {
                  this->opencl_comb_par.dev_id = 0;
                  this->opencl_sep_par.dev_id = 0;
                  this->opencl_comb_par.dev_type = CL_DEVICE_TYPE_GPU;
                  this->opencl_sep_par.dev_type = CL_DEVICE_TYPE_GPU;
                }
              else if((opt_value.compare("cpu") == 0) || (opt_value.compare("c") == 0))
                {
                  this->opencl_comb_par.dev_id = 0;
                  this->opencl_sep_par.dev_id = 0;
                  this->opencl_comb_par.dev_type = CL_DEVICE_TYPE_CPU;
                  this->opencl_sep_par.dev_type = CL_DEVICE_TYPE_CPU;
                }
              else
                {
                  ival = atoi(opt_value.c_str());
                  if(ival < 0)
                    {
                      if(error != NULL) *error << "Warning: Parameter \"device\" is not correct. Right values are (gpu|g|prefer-gpu) for prefer gpu (cpu|c|prefer-cpu) for prefer cpu (force-gpu) for force gpu (force-cpu) for force cpu or device id.\n";
                    }
                  else
                    {
                      this->opencl_comb_par.dev_id = ival;
                      this->opencl_sep_par.dev_id = ival;
                      this->opencl_comb_par.dev_type = CL_DEVICE_TYPE_ALL;
                      this->opencl_sep_par.dev_type = CL_DEVICE_TYPE_ALL;
                    }
                }
							type_count++;
						break; 
            case 5:
              this->opencl_comb_par.print_devices = true;
              this->opencl_sep_par.print_devices = true;
            break;
            case 6:
              ival = atoi(opt_value.c_str());
              if(ival < 0)
                {
                  if(error != NULL) *error << "Warning: Parameter \"kernel\" is out of range. Right range is <0,kernel_count).\n";
                }
              else
                {
                  this->kernel_id = ival;
                }
            break;
            case 7:
              if((opt_value.compare("gpu") == 0) || (opt_value.compare("g") == 0) || (opt_value.compare("1")))
                {
                  this->program_id = DET_PROGRAM_ID_GPU_PROC;
                }
              else
                {
                  if(error != NULL) *error << "Warning: Parameter \"program\" is out of range. Right range is <0,kernel_count).\n";
                }
            break;
            case 8:
              ival = atoi(opt_value.c_str());
              if(ival < 1)
                {
                  if(error != NULL) *error << "Warning: Parameter \"frame-count\" is out of range. Right range is <1,kernel_count).\n";
                }
              else
                {
                  this->frame_count = ival;
                }
            break;
            case 9:
              ival = atoi(opt_value.c_str());
              if(ival < 0)
              {
                if(error != NULL) *error << "Warning: Parameter \"camera\" is out of range. Right range is <0,inf).\n";
              }
              else
              {
                this->in_file = opt_value;
                this->cam_use = true;
              }
              break;
            case 10:
              ival = atoi(opt_value.c_str());
              if((ival < 1) || (ival > 8))
              {
                if(error != NULL) *error << "Warning: Parameter \"hor-pairs\" is out of range. Right range is <1,8>.\n";
              }
              else
              {
                this->opencl_sep_par.hor_pairs = ival;
                this->opencl_comb_par.hor_pairs = ival;
              }
              break;
            case 11:
              ival = atoi(opt_value.c_str());
              if(ival < 1)
              {
                if(error != NULL) *error << "Warning: Parameter \"vert-pairs\" is out of range. Right range is <1,inf).\n";
              }
              else
              {
                this->opencl_sep_par.vert_pairs = ival;
                this->opencl_comb_par.vert_pairs = ival;
              }
              break;
            case 12:
              ival = atoi(opt_value.c_str());
              if((ival < 1) || (ival > 1024))
              {
                if(error != NULL) *error << "Warning: Parameter \"hor-x\" is out of range. Right range is <1,1024>.\n";
              }
              else
              {
                this->opencl_sep_par.hor_sizes.x = ival;
              }
              break;
            case 13:
              ival = atoi(opt_value.c_str());
              if((ival < 1) || (ival > 1024))
              {
                if(error != NULL) *error << "Warning: Parameter \"hor-y\" is out of range. Right range is <1,1024>.\n";
              }
              else
              {
                this->opencl_sep_par.hor_sizes.y = ival;
              }
              break;
            case 14:
              ival = atoi(opt_value.c_str());
              if((ival < 1) || (ival > 1024))
              {
                if(error != NULL) *error << "Warning: Parameter \"vert-x\" is out of range. Right range is <1,1024>.\n";
              }
              else
              {
                this->opencl_sep_par.vert_sizes.x = ival;
              }
              break;
            case 15:
              ival = atoi(opt_value.c_str());
              if((ival < 1) || (ival > 1024))
              {
                if(error != NULL) *error << "Warning: Parameter \"vert-y\" is out of range. Right range is <1,1024>.\n";
              }
              else
              {
                this->opencl_sep_par.vert_sizes.y = ival;
              }
              break;
            case 16:
              ival = atoi(opt_value.c_str());
              if((ival < 1) || (ival > 1024))
              {
                if(error != NULL) *error << "Warning: Parameter \"comb-x\" is out of range. Right range is <1,1024>.\n";
              }
              else
              {
                this->opencl_comb_par.comb_sizes.x = ival;
              }
              break;
            case 17:
              ival = atoi(opt_value.c_str());
              if((ival < 1) || (ival > 1024))
              {
                if(error != NULL) *error << "Warning: Parameter \"comb-y\" is out of range. Right range is <1,1024>.\n";
              }
              else
              {
                this->opencl_comb_par.comb_sizes.y = ival;
              }
              break;
            case 18:
              this->opencl_sep_par.hor_kernel = opt_value;
              break;
            case 19:
              this->opencl_sep_par.vert_kernel = opt_value;
              break;
            case 20:
              this->opencl_comb_par.comb_kernel = opt_value;
              break;
            case 21:
              if((opt_value.compare("none") == 0) || (opt_value.compare("n") == 0))
                {
                  this->bench_type = CONF_BENCH_TYPE_NONE;
                }
              else if((opt_value.compare("csv") == 0) || (opt_value.compare("c") == 0))
                {
                  this->bench_type = CONF_BENCH_TYPE_CSV;
                }
              else
                {
                  if(error != NULL) *error << "Warning: Unknown type of parameter \"benchmark\". Supported types are (none|n) for non benchmarking mode (csv|c) for create csv file.\n";
                }
            break;
            case 22:
              this->opencl_comb_par.comb_hor_corners_proc = 1;
            break;
            case 23:
              ival = atoi(opt_value.c_str());
              if(ival < 1)
              {
                if(error != NULL) *error << "Warning: Parameter \"warp-size\" is out of range. Right range is <1,inf).\n";
              }
              else
              {
                this->opencl_sep_par.warp_size = ival;
                this->opencl_comb_par.warp_size = ival;
              }
              break;
            case 24:
              this->opencl_sep_par.atomic_warp = 0;
              this->opencl_comb_par.atomic_warp = 0;
            break;
            case 25:
              if((opt_value.compare("blaz_normal") == 0) || (opt_value.compare("bn") == 0))
                {
                  this->opencl_sep_par.hor_proc = WAVELET_PROC_TYPE_BLAZ_NORMAL;
                  this->opencl_comb_par.hor_proc = WAVELET_PROC_TYPE_BLAZ_NORMAL;
                }
              else if((opt_value.compare("blaz_register") == 0) || (opt_value.compare("br") == 0))
                {
                  this->opencl_sep_par.hor_proc = WAVELET_PROC_TYPE_BLAZ_REGISTER;
                  this->opencl_comb_par.hor_proc = WAVELET_PROC_TYPE_BLAZ_REGISTER;
                }
              else if((opt_value.compare("laan") == 0) || (opt_value.compare("l") == 0))
                {
                  this->opencl_sep_par.hor_proc = WAVELET_PROC_TYPE_LAAN;
                  this->opencl_comb_par.hor_proc = WAVELET_PROC_TYPE_LAAN;
                }
              else
                {
                  if(error != NULL) *error << "Warning: Wrong value for parameter \"hor-proc\".\n";
                }
            break;
            case 26:
              if((opt_value.compare("blaz_normal") == 0) || (opt_value.compare("bn") == 0))
                {
                  this->opencl_sep_par.vert_proc = WAVELET_PROC_TYPE_BLAZ_NORMAL;
                  this->opencl_comb_par.vert_proc = WAVELET_PROC_TYPE_BLAZ_NORMAL;
                }
              else if((opt_value.compare("blaz_register") == 0) || (opt_value.compare("br") == 0))
                {
                  this->opencl_sep_par.vert_proc = WAVELET_PROC_TYPE_BLAZ_REGISTER;
                  this->opencl_comb_par.vert_proc = WAVELET_PROC_TYPE_BLAZ_REGISTER;
                }
              else if((opt_value.compare("laan") == 0) || (opt_value.compare("l") == 0))
                {
                  this->opencl_sep_par.vert_proc = WAVELET_PROC_TYPE_LAAN;
                  this->opencl_comb_par.vert_proc = WAVELET_PROC_TYPE_LAAN;
                }
              else
                {
                  if(error != NULL) *error << "Warning: Wrong value for parameter \"vert-proc\".\n";
                }
            break;
            case 27:
              if(opt_value.compare("opencl") == 0)
                {
                  this->engine = ENGINE_TYPE_OPENCL_SEP;
                }
              else if(opt_value.compare("openmp") == 0)
                {
                  this->engine = ENGINE_TYPE_OPENMP_SEP;
                }
              else if(opt_value.compare("openclcomb") == 0)
                {
                  this->engine = ENGINE_TYPE_OPENCL_COMB;
                }
              else
                {
                  if(error != NULL) *error << "Warning: Wrong value of parameter \"engine-type\".\n";
                }
            break;
            case 28:
              if(opt_value.compare("transformSepSingle") == 0)
                {
                  this->openmp_sep_par.hor_func = OPENMP_SEP_FUNC_SINGLE;
                }
              else if(opt_value.compare("transformSepOpenmp") == 0)
                {
                  this->openmp_sep_par.hor_func = OPENMP_SEP_FUNC_OPENMP;
                }
              else if(opt_value.compare("transformSepOpenmpCopy") == 0)
                {
                  this->openmp_sep_par.hor_func = OPENMP_SEP_FUNC_OPENMP_COPY;
                }
              else if(opt_value.compare("transformSepOpenmpSimd") == 0)
                {
                  this->openmp_sep_par.hor_func = OPENMP_SEP_FUNC_OPENMP_SIMD;
                }
              else if(opt_value.compare("transformSepOpenmpCopySimd") == 0)
                {
                  this->openmp_sep_par.hor_func = OPENMP_SEP_FUNC_OPENMP_COPY_SIMD;
                }
              else
                {
                  if(error != NULL) *error << "Warning: Wrong value of parameter \"hor-func\".\n";
                }
            break;
            case 29:
              if(opt_value.compare("transformSepSingle") == 0)
                {
                  this->openmp_sep_par.vert_func = OPENMP_SEP_FUNC_SINGLE;
                }
              else if(opt_value.compare("transformSepOpenmp") == 0)
                {
                  this->openmp_sep_par.vert_func = OPENMP_SEP_FUNC_OPENMP;
                }
              else if(opt_value.compare("transformSepOpenmpCopy") == 0)
                {
                  this->openmp_sep_par.vert_func = OPENMP_SEP_FUNC_OPENMP_COPY;
                }
              else if(opt_value.compare("transformSepOpenmpSimd") == 0)
                {
                  this->openmp_sep_par.vert_func = OPENMP_SEP_FUNC_OPENMP_SIMD;
                }
              else if(opt_value.compare("transformSepOpenmpCopySimd") == 0)
                {
                  this->openmp_sep_par.vert_func = OPENMP_SEP_FUNC_OPENMP_COPY_SIMD;
                }
              else
                {
                  if(error != NULL) *error << "Warning: Wrong value of parameter \"vert-func\".\n";
                }
            break;
            case 30:
              this->opencl_comb_par.interlaced = true;
              this->opencl_sep_par.interlaced = true;
			  this->openmp_sep_par.interlaced = true;
            break;
			case 31:
				ival = atoi(opt_value.c_str());
				if(ival < 1)
				{
					if(error != NULL) *error << "Warning: Parameter \"depth\" is out of range. Right range is <1,inf>.\n";
				}
				else
				{
					this->opencl_comb_par.max_depth = ival;
					this->opencl_sep_par.max_depth = ival;
					this->openmp_sep_par.max_depth = ival;
				}
			break;
						case PARSE_ARG_NOT_FOUND:
							if(error != NULL) *error << "Warning: Parameter \"" << argv[i] << "\" not found.\n";
					}
			}
		if(type_count > 1)
			{
				if(error != NULL) *error << "Warning: Cannot have more than 1 device type specifiers. Uses last one.\n";
			}
    if((this->in_file.length() == 0) && (!this->cam_use))
      {
        if(error != NULL) *error << "Warning: No input file or camera specified. Default camera used.\n";
        this->cam_use = true;
        this->in_file = "0";
      }
  }
