#include "WaveletOpenclParamComb.h"

WaveletOpenclParamComb::WaveletOpenclParamComb() : WaveletOpenclParam(ENGINE_TYPE_OPENCL_COMB)
  {
    this->clear();
  }
/**
  * Set initial undefined value.
  */
void WaveletOpenclParamComb::clear()
  {
    this->WaveletOpenclParam::clear();
    this->comb_sizes = proc_dim(32,8);
    this->comb_kernel = std::string("wavelet_block_in_2step_20mac");
    this->comb_hor_corners_proc = 0;
  }
/**
  * Get program build parameters.
	* @param block_size threads in block count
	* @param stat_creating statistics creating type
	* @return parameters
  */
std::string WaveletOpenclParamComb::createBuildParam()
  {
    std::ostringstream str;

    str << this->WaveletOpenclParam::createBuildParam()
        << " -D COMB_BLOCK_SIZE=" << this->comb_sizes.count()
        << " -D COMB_BLOCK_SIZE_X=" << this->comb_sizes.x
        << " -D COMB_HOR_CORNERS_PROC=" << this->comb_hor_corners_proc
        << " -D " << this->comb_kernel;
    #if DEBUG_LEVEL == DEBUG_ALL
      this->printDebug();
    #endif
    return str.str();
  }
/**
  * Print debug information
  */
void WaveletOpenclParamComb::printDebug()
  {
    fprintf(stderr, " comb kernel name:        %s\n"
                    " comb kernel local sizes: %dx%d\n"
                    " comb corners process:    %s\n", this->comb_kernel.c_str(), this->comb_sizes.x, this->comb_sizes.y, (this->comb_hor_corners_proc == 1) ? "yes" : "no");
  }